# demo.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Demonstration of using the GoPro package."""

import sys
import time
import logging
import argparse
import threading
from json import dumps
from pathlib import Path
from typing import Tuple

from rich import traceback
from rich.logging import RichHandler
from rich.console import Console

from open_gopro import GoPro, params, QueryCmdId
from open_gopro.util import launch_vlc

logger = logging.getLogger(__name__)
traceback.install()  # Enable exception tracebacks in rich logger
console = Console()  # rich consoler printer


def process_async_events(gopro: GoPro) -> None:
    """Separate thread to continuously get and print asynchronous events from the GoPro instance.

    Args:
        gopro (GoPro): instance to get updates from
    """
    while True:
        # Block until we receive one
        update = gopro.get_update()

        # Print pushed updates (they're either setting capability or setting / status value updates)
        description = "capabilities are" if update.cmd is QueryCmdId.SETTING_CAPABILITY_PUSH else "value is"
        for item in update:
            console.print(f"{item} {description} now: {dumps(update[item])}")


def main() -> None:
    """Main function."""
    identifier, log_location, vlc_location = parse_arguments()
    setup_logging(log_location)

    try:
        with GoPro(identifier) as gopro:
            # Dump services to CSV
            gopro.services_as_csv()

            # Start a thread to handle asynchronous notifications
            threading.Thread(target=process_async_events, args=(gopro,), daemon=True).start()

            # Now we only want errors and above since we're going to be doing a lot of printing below
            for handler in logger.handlers:
                if isinstance(handler, RichHandler):
                    handler.setLevel(logging.ERROR)

            with console.status("[bold green]Ensuring the shutter is off..."):
                assert gopro.ble_command.set_shutter(params.Shutter.OFF).is_ok

            with console.status("[bold green]Ensuring turbo mode is off..."):
                assert gopro.ble_command.set_turbo_mode(False).is_ok

            with console.status("[bold green]Getting all statuses via BLE..."):
                for status, value in gopro.ble_command.get_camera_status().items():
                    console.print(f"{status} : {value}")

            with console.status("[bold green]Getting settings and statuses via WiFi..."):
                for item, value in gopro.wifi_command.get_camera_state().items():
                    console.print(f"{item} : {value}")

            with console.status("[bold green]Getting Open GoPro info..."):
                assert gopro.keep_alive()
                assert gopro.ble_command.set_third_party_client_info().is_ok
                version = gopro.ble_command.get_third_party_api_version()
                assert version.is_ok
                console.print(f"Version is {version['major']}.{version['minor']}")

            with console.status("[bold green]Setting Cinematic preset..."):
                assert gopro.ble_command.load_preset(params.Preset.CINEMATIC).is_ok

            with console.status("[bold green]Getting some individual values via BLE..."):
                resolution = gopro.ble_setting.resolution.get_value().flatten
                console.print(f"Current resolution is {resolution}")
                fov = gopro.ble_setting.video_field_of_view.get_value().flatten
                console.print(f"Current video FOV is {fov}")
                encoding = gopro.ble_status.encoding_active.get_value().flatten
                console.print(f"Current encoding status is {encoding}")

            with console.status("[bold green]Getting some capabilities via BLE..."):
                resolution_caps = gopro.ble_setting.resolution.get_capabilities_values().flatten
                console.print(f"Current resolution capabilities are: {dumps(resolution_caps)}")
                fov_caps = gopro.ble_setting.video_field_of_view.get_capabilities_values().flatten
                console.print(f"Current video FOV capabilities are: {dumps(fov_caps)}")

            with console.status("[bold green]Registering for push notifications..."):
                assert gopro.ble_setting.resolution.set(params.Resolution.RES_1080).is_ok
                fps_caps_update = gopro.ble_setting.fps.register_value_update().flatten
                console.print(f"New fps capabilities are: {dumps(fps_caps_update)}")
                assert gopro.ble_setting.resolution.register_value_update().is_ok
                assert gopro.ble_setting.resolution.register_capability_update().is_ok
                assert gopro.ble_setting.fps.register_capability_update().is_ok
                assert gopro.ble_setting.video_field_of_view.register_value_update().is_ok
                assert gopro.ble_setting.video_field_of_view.register_capability_update().is_ok
                assert gopro.ble_status.encoding_active.register_value_update().is_ok

            with console.status("[bold green]Taking a photo..."):
                assert gopro.ble_command.load_preset(params.Preset.PHOTO).is_ok
                assert gopro.ble_command.set_shutter(params.Shutter.ON).is_ok

            with console.status("[bold green]Taking a video..."):
                assert gopro.ble_command.load_preset(params.Preset.CINEMATIC).is_ok
                assert gopro.ble_command.set_shutter(params.Shutter.ON).is_ok
                time.sleep(2)  # Take a 2 second video
                assert gopro.ble_command.set_shutter(params.Shutter.OFF).is_ok

            with console.status("[bold green]Testing some capability pushes"):
                assert gopro.ble_setting.resolution.set(params.Resolution.RES_1080).is_ok
                assert gopro.ble_setting.fps.set(params.FPS.FPS_30).is_ok
                assert gopro.ble_setting.fps.set(params.FPS.FPS_240).is_ok
                assert gopro.ble_setting.resolution.get_capabilities_values().is_ok
                assert gopro.ble_setting.video_field_of_view.set(params.FieldOfView.LINEAR).is_ok
                assert gopro.ble_setting.video_field_of_view.set(params.FieldOfView.NARROW).is_ok
                assert gopro.ble_setting.video_field_of_view.set(params.FieldOfView.WIDE).is_ok
                # We expect this to fail as it is not valid with the current resolution
                assert not gopro.ble_setting.video_field_of_view.set(params.FieldOfView.SUPERVIEW).is_ok
                # Cycle through resolutions. This will cause other settings (and their capabilities) to update
                for resolution in resolution_caps:
                    assert gopro.ble_setting.resolution.set(resolution).is_ok

            with console.status("[bold green]Get media list and download some media..."):
                assert gopro.wifi_command.set_preset(params.Preset.PHOTO).is_ok
                media_list = gopro.wifi_command.get_media_list()["media"][0]["fs"]
                # Find a picture and download it
                picture = ""
                for file in [x["n"] for x in media_list]:
                    if file.lower().endswith(".jpg"):
                        picture = file
                        gopro.wifi_command.download_file(camera_file=picture)
                        break
                # Find a video and download it
                video = ""
                for file in [x["n"] for x in media_list]:
                    if file.lower().endswith(".mp4"):
                        video = file
                        gopro.wifi_command.download_file(camera_file=video)
                        break

            with console.status("[bold green]Getting media information..."):
                assert gopro.wifi_command.get_media_info(video).is_ok
                assert gopro.wifi_command.get_media_info(picture).is_ok
                gopro.wifi_command.get_gpmf_data(camera_file=picture)
                gopro.wifi_command.get_screennail(camera_file=video)
                gopro.wifi_command.get_telemetry(camera_file=video)
                gopro.wifi_command.get_thumbnail(camera_file=picture)
                assert gopro.wifi_command.get_preset_status().is_ok

            with console.status("[bold green]Cycling through resolutions via WiFi..."):
                assert gopro.wifi_command.set_preset(params.Preset.CINEMATIC).is_ok
                assert gopro.wifi_setting.resolution.set(params.Resolution.RES_1080).is_ok
                assert gopro.wifi_setting.resolution.set(params.Resolution.RES_1440).is_ok
                assert gopro.wifi_setting.resolution.set(params.Resolution.RES_2_7k).is_ok
                assert gopro.wifi_setting.resolution.set(params.Resolution.RES_5k).is_ok
                assert gopro.wifi_setting.resolution.set(params.Resolution.RES_1080).is_ok

            with console.status("[bold green]Starting the preview stream..."):
                gopro.wifi_command.stop_preview_stream()
                gopro.wifi_command.start_preview_stream()

            console.print("Preview stream is enabled. View with a UDP server at udp://@:8554.")
            console.print("Send keyboard interrupt to exit.")

            threading.Thread(target=launch_vlc, args=(vlc_location,), daemon=True).start()

            while True:
                time.sleep(0.2)

    except Exception as e:  # pylint: disable=broad-except
        logger.error(repr(e))
        sys.exit(-1)
    except KeyboardInterrupt:
        logger.warning("Received keyboard interrupt. Shutting down...")
    finally:
        console.print("Exiting...")
        sys.exit(0)


def setup_logging(log_location: Path) -> None:
    """Configure logging to file and Rich console logging

    Args:
        log_location (Path): location to configure for the file handler
    """
    # Logging to file with
    fh = logging.FileHandler(f"{log_location}", mode="w")
    file_formatter = logging.Formatter(
        fmt="%(threadName)13s: %(name)40s:%(lineno)5d %(asctime)s.%(msecs)03d %(levelname)-8s | %(message)s",
        datefmt="%H:%M:%S",
    )
    fh.setFormatter(file_formatter)
    fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)

    # Use Rich for colorful console logging
    sh = RichHandler(rich_tracebacks=True, enable_link_path=True, show_time=False)
    stream_formatter = logging.Formatter("%(asctime)s.%(msecs)03d %(message)s", datefmt="%H:%M:%S")
    sh.setFormatter(stream_formatter)
    sh.setLevel(logging.INFO)
    logger.addHandler(sh)
    logger.setLevel(logging.DEBUG)

    # Enable / disable logging in modules
    for (module, level) in [
        ("open_gopro.gopro", logging.DEBUG),
        ("open_gopro.ble_commands", logging.DEBUG),
        ("open_gopro.ble_controller", logging.DEBUG),
        ("open_gopro.wifi_commands", logging.DEBUG),
        ("open_gopro.wifi_controller", logging.DEBUG),
        ("open_gopro.responses", logging.DEBUG),
        ("open_gopro.util", logging.DEBUG),
        ("bleak", logging.DEBUG),
        ("bleak.backends.bluezdbus.client", logging.DEBUG),
        ("bleak.backends.corebluetooth.client", logging.DEBUG),
        ("bleak.backends.dotnet.client", logging.DEBUG),
    ]:
        log = logging.getLogger(module)
        log.setLevel(level)
        log.addHandler(fh)
        log.addHandler(sh)


def parse_arguments() -> Tuple[str, Path, Path]:
    """Parse command line arguments

    Returns:
        Tuple[str, Path, Path]: (identifier, path to save log, path to VLC)
    """
    parser = argparse.ArgumentParser(
        description="Connect to a GoPro camera via BLE and Wifi and do some things."
    )
    parser.add_argument(
        "-i",
        "--identifier",
        type=str,
        help="Last 4 digits of GoPro serial number, which is the last 4 digits of the default camera SSID. \
            If not used, first discovered GoPro will be connected to",
        default=None,
    )
    parser.add_argument(
        "-l",
        "--log",
        type=Path,
        help="Location to store detailed log",
        default="gopro_demo.log",
    )
    parser.add_argument(
        "-v",
        "--vlc",
        type=Path,
        help="VLC location. If not set, the location will attempt to be automatically discovered.",
        default=None,
    )
    args = parser.parse_args()

    return args.identifier, args.log, args.vlc


if __name__ == "__main__":
    main()



# __init__.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""All Protobuf defnitions."""

from open_gopro.proto.turbo_transfer_pb import RequestSetTurboActive
from open_gopro.proto.response_generic_pb import ResponseGeneric
from open_gopro.proto.request_get_preset_status_pb import RequestGetPresetStatus
from open_gopro.proto.preset_status_pb import NotifyPresetStatus

__all__ = ["RequestSetTurboActive", "ResponseGeneric", "RequestGetPresetStatus", "NotifyPresetStatus"]



# preset_status_pb.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: preset_status.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto


class EnumPresetGroup(betterproto.Enum):
    PRESET_GROUP_ID_VIDEO = 1000
    PRESET_GROUP_ID_PHOTO = 1001
    PRESET_GROUP_ID_TIMELAPSE = 1002
    PRESET_GROUP_ID_VIDEO_DUAL_LENS = 1003
    PRESET_GROUP_ID_PHOTO_DUAL_LENS = 1004
    PRESET_GROUP_ID_TIMELAPSE_DUAL_LENS = 1005
    PRESET_GROUP_ID_SPECIAL = 1006


class EnumFlatMode(betterproto.Enum):
    FLAT_MODE_UNKNOWN = -1
    FLAT_MODE_PLAYBACK = 4
    FLAT_MODE_SETUP = 5
    FLAT_MODE_VIDEO = 12
    FLAT_MODE_TIME_LAPSE_VIDEO = 13
    FLAT_MODE_LOOPING = 15
    FLAT_MODE_PHOTO = 17
    FLAT_MODE_PHOTO_NIGHT = 18
    FLAT_MODE_PHOTO_BURST = 19
    FLAT_MODE_TIME_LAPSE_PHOTO = 20
    FLAT_MODE_NIGHT_LAPSE_PHOTO = 21
    FLAT_MODE_BROADCAST_BROADCAST = 23
    FLAT_MODE_TIME_WARP_VIDEO = 24
    FLAT_MODE_LIVE_BURST = 25
    FLAT_MODE_NIGHT_LAPSE_VIDEO = 26
    FLAT_MODE_SLOMO = 27


class EnumPresetTitle(betterproto.Enum):
    PRESET_TITLE_ACTIVITY = 0
    PRESET_TITLE_STANDARD = 1
    PRESET_TITLE_CINEMATIC = 2
    PRESET_TITLE_PHOTO = 3
    PRESET_TITLE_LIVE_BURST = 4
    PRESET_TITLE_BURST = 5
    PRESET_TITLE_NIGHT = 6
    PRESET_TITLE_TIME_WARP = 7
    PRESET_TITLE_TIME_LAPSE = 8
    PRESET_TITLE_NIGHT_LAPSE = 9
    PRESET_TITLE_VIDEO = 10
    PRESET_TITLE_SLOMO = 11
    PRESET_TITLE_360_VIDEO = 12
    PRESET_TITLE_PHOTO_2 = 13
    PRESET_TITLE_PANORAMA = 14
    PRESET_TITLE_360_PHOTO = 15
    PRESET_TITLE_TIME_WARP_2 = 16
    PRESET_TITLE_360_TIME_WARP = 17
    PRESET_TITLE_CUSTOM = 18
    PRESET_TITLE_AIR = 19
    PRESET_TITLE_BIKE = 20
    PRESET_TITLE_EPIC = 21
    PRESET_TITLE_INDOOR = 22
    PRESET_TITLE_MOTOR = 23
    PRESET_TITLE_MOUNTED = 24
    PRESET_TITLE_OUTDOOR = 25
    PRESET_TITLE_POV = 26
    PRESET_TITLE_SELFIE = 27
    PRESET_TITLE_SKATE = 28
    PRESET_TITLE_SNOW = 29
    PRESET_TITLE_TRAIL = 30
    PRESET_TITLE_TRAVEL = 31
    PRESET_TITLE_WATER = 32
    PRESET_TITLE_LOOPING = 33
    # Reserved 34 - 50 for custom presets.
    PRESET_TITLE_360_TIMELAPSE = 51
    PRESET_TITLE_360_NIGHT_LAPSE = 52
    PRESET_TITLE_360_NIGHT_PHOTO = 53
    PRESET_TITLE_PANO_TIME_LAPSE = 54
    PRESET_TITLE_MAX_VIDEO = 55
    PRESET_TITLE_MAX_PHOTO = 56
    PRESET_TITLE_MAX_TIMEWARP = 57
    PRESET_TITLE_MAX = 58


class EnumPresetIcon(betterproto.Enum):
    PRESET_ICON_VIDEO = 0
    PRESET_ICON_ACTIVITY = 1
    PRESET_ICON_CINEMATIC = 2
    PRESET_ICON_PHOTO = 3
    PRESET_ICON_LIVE_BURST = 4
    PRESET_ICON_BURST = 5
    PRESET_ICON_PHOTO_NIGHT = 6
    PRESET_ICON_TIMEWARP = 7
    PRESET_ICON_TIMELAPSE = 8
    PRESET_ICON_NIGHTLAPSE = 9
    PRESET_ICON_SNAIL = 10
    PRESET_ICON_VIDEO_2 = 11
    PRESET_ICON_360_VIDEO = 12
    PRESET_ICON_PHOTO_2 = 13
    PRESET_ICON_PANORAMA = 14
    PRESET_ICON_BURST_2 = 15
    PRESET_ICON_TIMEWARP_2 = 16
    PRESET_ICON_TIMELAPSE_2 = 17
    PRESET_ICON_CUSTOM = 18
    PRESET_ICON_AIR = 19
    PRESET_ICON_BIKE = 20
    PRESET_ICON_EPIC = 21
    PRESET_ICON_INDOOR = 22
    PRESET_ICON_MOTOR = 23
    PRESET_ICON_MOUNTED = 24
    PRESET_ICON_OUTDOOR = 25
    PRESET_ICON_POV = 26
    PRESET_ICON_SELFIE = 27
    PRESET_ICON_SKATE = 28
    PRESET_ICON_SNOW = 29
    PRESET_ICON_TRAIL = 30
    PRESET_ICON_TRAVEL = 31
    PRESET_ICON_WATER = 32
    PRESET_ICON_LOOPING = 33
    # Add icons below for new presets starting from 51
    PRESET_ICON_MAX_VIDEO = 55
    PRESET_ICON_MAX_PHOTO = 56
    PRESET_ICON_MAX_TIMEWARP = 57
    PRESET_ICON_TIMELAPSE_PHOTO = 1000
    PRESET_ICON_NIGHTLAPSE_PHOTO = 1001
    PRESET_ICON_MAX = 1002


@dataclass
class NotifyPresetStatus(betterproto.Message):
    preset_group_array: List["PresetGroup"] = betterproto.message_field(1)


@dataclass
class PresetGroup(betterproto.Message):
    id: "EnumPresetGroup" = betterproto.enum_field(1)
    preset_array: List["Preset"] = betterproto.message_field(2)
    can_add_preset: bool = betterproto.bool_field(3)


@dataclass
class Preset(betterproto.Message):
    id: int = betterproto.int32_field(1)
    mode: "EnumFlatMode" = betterproto.enum_field(2)
    title_id: "EnumPresetTitle" = betterproto.enum_field(3)
    title_number: int = betterproto.int32_field(4)
    user_defined: bool = betterproto.bool_field(5)
    icon: "EnumPresetIcon" = betterproto.enum_field(6)
    setting_array: List["PresetSetting"] = betterproto.message_field(7)
    is_modified: bool = betterproto.bool_field(8)


@dataclass
class PresetSetting(betterproto.Message):
    id: int = betterproto.int32_field(1)
    value: int = betterproto.int32_field(2)
    is_caption: bool = betterproto.bool_field(3)



# request_get_preset_status_pb.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: request_get_preset_status.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto


class EnumRegisterPresetStatus(betterproto.Enum):
    REGISTER_PRESET_STATUS_PRESET = 1
    REGISTER_PRESET_STATUS_PRESET_GROUP_ARRAY = 2


@dataclass
class RequestGetPresetStatus(betterproto.Message):
    register_preset_status: List["EnumRegisterPresetStatus"] = betterproto.enum_field(1)
    unregister_preset_status: List["EnumRegisterPresetStatus"] = betterproto.enum_field(2)



# constants.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Constant numbers shared across the GoPro module."""

import enum
from typing import Union, Tuple

import construct

GOPRO_BASE_UUID = "b5f9{}-aa8d-11e3-9046-0002a5d5c51b"


class ErrorCode(enum.Enum):
    """Status Codes."""

    SUCCESS = 0
    ERROR = 1
    INVALID_PARAM = 2


class UUID(enum.Enum):
    """BLE UUID."""

    # Generic Attribute Service
    S_GENERIC_ATT = "00001801-0000-1000-8000-00805f9b34fb"

    # Generic Access Service
    S_GENERIC_ACCESS = "00001800-0000-1000-8000-00805f9b34fb"
    ACC_DEVICE_NAME = "00002a00-0000-1000-8000-00805f9b34fb"
    ACC_APPEARANCE = "00002a01-0000-1000-8000-00805f9b34fb"
    ACC_PREF_CONN_PARAMS = "00002a04-0000-1000-8000-00805f9b34fb"
    ACC_CENTRAL_ADDR_RES = "00002aa6-0000-1000-8000-00805f9b34fb"

    # Tx Power
    S_TX_POWER = "00001804-0000-1000-8000-00805f9b34fb"
    TX_POWER_LEVEL = "00002a07-0000-1000-8000-00805f9b34fb"

    # Battery Service
    S_BATTERY = "0000180f-0000-1000-8000-00805f9b34fb"
    BATT_LEVEL = "00002a19-0000-1000-8000-00805f9b34fb"

    # Device Information Service
    S_DEV_INFO = "0000180a-0000-1000-8000-00805f9b34fb"
    INF_MAN_NAME = "00002a29-0000-1000-8000-00805f9b34fb"
    INF_MODEL_NUM = "00002a24-0000-1000-8000-00805f9b34fb"
    INF_SERIAL_NUM = "00002a25-0000-1000-8000-00805f9b34fb"
    INF_FW_REV = "00002a26-0000-1000-8000-00805f9b34fb"
    INF_HW_REV = "00002a27-0000-1000-8000-00805f9b34fb"
    INF_SW_REV = "00002a28-0000-1000-8000-00805f9b34fb"
    INF_SYS_ID = "00002a23-0000-1000-8000-00805f9b34fb"
    INF_CERT_DATA = "00002a2a-0000-1000-8000-00805f9b34fb"
    INF_PNP_ID = "00002a50-0000-1000-8000-00805f9b34fb"

    # GoPro Wifi Access Point Service
    S_WIFI_ACCESS_POINT = GOPRO_BASE_UUID.format("0001")
    WAP_SSID = GOPRO_BASE_UUID.format("0002")
    WAP_PASSWORD = GOPRO_BASE_UUID.format("0003")
    WAP_POWER = GOPRO_BASE_UUID.format("0004")
    WAP_STATE = GOPRO_BASE_UUID.format("0005")
    WAP_CSI_PASSWORD = GOPRO_BASE_UUID.format("0006")

    # GoPro Control & Query Service
    S_CONTROL_QUERY = "0000fea6-0000-1000-8000-00805f9b34fb"
    CQ_COMMAND = GOPRO_BASE_UUID.format("0072")
    CQ_COMMAND_RESP = GOPRO_BASE_UUID.format("0073")
    CQ_SETTINGS = GOPRO_BASE_UUID.format("0074")
    CQ_SETTINGS_RESP = GOPRO_BASE_UUID.format("0075")
    CQ_QUERY = GOPRO_BASE_UUID.format("0076")
    CQ_QUERY_RESP = GOPRO_BASE_UUID.format("0077")
    CQ_SENSOR = GOPRO_BASE_UUID.format("0078")
    CQ_SENSOR_RESP = GOPRO_BASE_UUID.format("0079")

    # GoPro Camera Management Service
    S_CAMERA_MANAGEMENT = GOPRO_BASE_UUID.format("0090")
    CM_NET_MGMT_COMM = GOPRO_BASE_UUID.format("0091")
    CN_NET_MGMT_RESP = GOPRO_BASE_UUID.format("0092")

    # Unknown
    S_UNKNOWN = GOPRO_BASE_UUID.format("0080")
    INTERNAL_81 = GOPRO_BASE_UUID.format("0081")
    INTERNAL_82 = GOPRO_BASE_UUID.format("0082")
    INTERNAL_83 = GOPRO_BASE_UUID.format("0083")
    INTERNAL_84 = GOPRO_BASE_UUID.format("0084")


class CmdId(enum.Enum):
    """Command ID's that are written to UUID.CQ_COMMAND."""

    SET_SHUTTER = 0x01
    SET_PAIRING_COMPLETE = 0x03
    GET_CAMERA_STATUS = 0x13
    SET_WIFI = 0x17
    GET_SETTINGS_JSON = 0x3B
    GET_HW_INFO = 0x3C
    LOAD_PRESET = 0x40
    SET_THIRD_PARTY_CLIENT_INFO = 0x50
    GET_THIRD_PARTY_API_VERSION = 0x51
    SET_TURBO_MODE = 0xF1
    GET_PRESET_STATUS = 0xF5


class ActionId(enum.Enum):
    """Action ID's that identify a protobuf command."""

    SET_TURBO_MODE = 0x6B
    GET_PRESET_STATUS = 0x02


class SettingId(enum.Enum):
    """Setting ID's that identify settings and are written to UUID.CQ_SETTINGS."""

    RESOLUTION = 2
    FPS = 3
    INTERNAL_5 = 5
    INTERNAL_6 = 6
    INTERNAL_13 = 13
    INTERNAL_19 = 19
    INTERNAL_24 = 24
    INTERNAL_30 = 30
    INTERNAL_31 = 31
    INTERNAL_32 = 32
    INTERNAL_37 = 37
    INTERNAL_41 = 41
    INTERNAL_42 = 42
    INTERNAL_43 = 43
    INTERNAL_44 = 44
    INTERNAL_45 = 45
    INTERNAL_47 = 47
    INTERNAL_48 = 48
    INTERNAL_54 = 54
    INTERNAL_59 = 59
    INTERNAL_60 = 60
    INTERNAL_61 = 61
    INTERNAL_62 = 62
    INTERNAL_64 = 64
    INTERNAL_65 = 65
    INTERNAL_66 = 66
    INTERNAL_67 = 67
    INTERNAL_75 = 75
    INTERNAL_76 = 76
    INTERNAL_79 = 79
    INTERNAL_83 = 83
    INTERNAL_84 = 84
    INTERNAL_85 = 85
    INTERNAL_86 = 86
    INTERNAL_87 = 87
    INTERNAL_88 = 88
    LED = 91
    INTERNAL_96 = 96
    INTERNAL_102 = 102
    INTERNAL_103 = 103
    INTERNAL_104 = 104
    INTERNAL_105 = 105
    INTERNAL_106 = 106
    INTERNAL_111 = 111
    INTERNAL_112 = 112
    INTERNAL_114 = 114
    INTERNAL_115 = 115
    INTERNAL_116 = 116
    INTERNAL_117 = 117
    INTERNAL_118 = 118
    VIDEO_FOV = 121
    PHOTO_FOV = 122
    MULTI_SHOT_FOV = 123
    INTERNAL_124 = 124
    INTERNAL_125 = 125
    INTERNAL_126 = 126
    INTERNAL_128 = 128
    SHORTCUT_LOWER_LEFT = 129
    SHORTCUT_LOWER_RIGHT = 130
    SHORTCUT_UPPER_LEFT = 131
    SHORTCUT_UPPER_RIGHT = 132
    INTERNAL_133 = 133
    INTERNAL_134 = 134
    INTERNAL_135 = 135
    INTERNAL_139 = 139
    FLATMODE = 144
    INTERNAL_145 = 145
    INTERNAL_146 = 146
    INTERNAL_147 = 147
    INTERNAL_148 = 148
    INTERNAL_149 = 149
    INTERNAL_153 = 153
    INTERNAL_154 = 154
    INTERNAL_155 = 155
    INTERNAL_156 = 156
    INTERNAL_157 = 157
    INTERNAL_158 = 158
    INTERNAL_159 = 159
    INTERNAL_160 = 160
    INTERNAL_161 = 161
    MAX_LENS_MOD = 162
    INTERNAL_163 = 163
    INTERNAL_164 = 164
    INTERNAL_165 = 165
    INTERNAL_166 = 166
    INTERNAL_167 = 167
    INTERNAL_168 = 168
    INTERNAL_169 = 169
    INVALID_FOR_TESTING = 0xFF


class QueryCmdId(enum.Enum):
    """Command ID that is written to UUID.CQ_QUERY."""

    GET_SETTING_VAL = 0x12
    GET_STATUS_VAL = 0x13
    GET_SETTING_NAME = 0x22
    GET_CAPABILITIES_VAL = 0x32
    GET_CAPABILITIES_NAME = 0x42
    REG_SETTING_VAL_UPDATE = 0x52
    REG_STATUS_VAL_UPDATE = 0x53
    REG_CAPABILITIES_UPDATE = 0x62
    UNREG_SETTING_VAL_UPDATE = 0x72
    UNREG_STATUS_VAL_UPDATE = 0x73
    UNREG_CAPABILITIES_UPDATE = 0x82
    SETTING_VAL_PUSH = 0x92
    STATUS_VAL_PUSH = 0x93
    SETTING_CAPABILITY_PUSH = 0xA2
    INVALID_FOR_TESTING = 0xFF


class StatusId(enum.Enum):
    """Status ID to identify statuses sent to UUID.CQ_QUERY or received from UUID.CQ_QUERY_RESP."""

    BATT_PRESENT = 1
    BATT_LEVEL = 2
    EXT_BATT_PRESENT = 3
    EXT_BATT_LEVEL = 4
    SYSTEM_HOT = 6
    SYSTEM_BUSY = 8
    QUICK_CAPTURE = 9
    ENCODING = 10
    LCD_LOCK_ACTIVE = 11
    VIDEO_PROGRESS = 13
    INTERNAL_14 = 14
    WIRELESS_ENABLED = 17
    PAIR_STATE = 19
    PAIR_TYPE = 20
    PAIR_TIME = 21
    WAP_SCAN_STATE = 22
    WAP_SCAN_TIME = 23
    WAP_PROV_STAT = 24
    REMOTE_CTRL_VER = 26
    REMOTE_CTRL_CONN = 27
    PAIR_STATE2 = 28
    WLAN_SSID = 29
    AP_SSID = 30
    APP_COUNT = 31
    PREVIEW_ENABLED = 32
    SD_STATUS = 33
    PHOTOS_REM = 34
    VIDEO_REM = 35
    NUM_GROUP_PHOTO = 36
    NUM_GROUP_VIDEO = 37
    NUM_TOTAL_PHOTO = 38
    NUM_TOTAL_VIDEO = 39
    DATE_TIME = 40
    OTA_STAT = 41
    DOWNLAD_CANCEL_PEND = 42
    MODE_GROUP = 43
    LOCATE_ACTIVE = 45
    INTERNAL_46 = 46
    INTERNAL_47 = 47
    INTERNAL_48 = 48
    MULTI_COUNT_DOWN = 49
    SPACE_REM = 54
    STREAMING_SUPP = 55
    WIFI_BARS = 56
    CURRENT_TIME_MS = 57
    NUM_HILIGHTS = 58
    LAST_HILIGHT = 59
    NEXT_POLL = 60
    ANALYTICS_RDY = 61
    ANALYTICS_SIZE = 62
    IN_CONTEXT_MENU = 63
    TIMELAPSE_REM = 64
    EXPOSURE_TYPE = 65
    EXPOSURE_X = 66
    EXPOSURE_Y = 67
    GPS_STAT = 68
    AP_STATE = 69
    INT_BATT_PER = 70
    ACC_MIC_STAT = 74
    DIGITAL_ZOOM = 75
    WIRELESS_BAND = 76
    DIG_ZOOM_ACTIVE = 77
    MOBILE_VIDEO = 78
    FIRST_TIME = 79
    SEC_SD_STAT = 80
    BAND_5GHZ_AVAIL = 81
    SYSTEM_READY = 82
    BATT_OK_OTA = 83
    CAPTURE_DELAY = 84
    VIDEO_LOW_TEMP = 85
    ORIENTATION = 86
    THERMAL_MIT_MODE = 87
    ZOOM_ENCODING = 88
    FLATMODE_ID = 89
    INTERNAL_90 = 90
    LOGS_READY = 91
    TIMEWARP_1X_ACTIVE = 92
    VIDEO_PRESETS = 93
    PHOTO_PRESETS = 94
    TIMELAPSE_PRESETS = 95
    PRESETS_GROUP = 96
    ACTIVE_PRESET = 97
    PRESET_MODIFIED = 98
    LIVE_BURST_REM = 99
    LIVE_BURST_TOTAL = 100
    CAPT_DELAY_ACTIVE = 101
    MEDIA_MOD_MIC_STAT = 102
    TIMEWARP_SPEED_RAMP = 103
    LINUX_CORE_ACTIVE = 104
    CAMERA_LENS_TYPE = 105
    VIDEO_HINDSIGHT = 106
    SCHEDULED_PRESET = 107
    SCHEDULED_CAPTURE = 108
    CREATING_PRESET = 109
    MEDIA_MOD_STAT = 110
    TURBO_MODE = 113


ProducerType = Tuple[QueryCmdId, Union[SettingId, StatusId]]
"""Types that can be registered for."""

CmdType = Union[CmdId, QueryCmdId, ActionId]
"""Types that identify a command."""

ResponseType = Union[CmdType, StatusId, SettingId, UUID, str, construct.Enum]
"""Types that are used to identify a response."""



# interfaces.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Interfaces that must be defined outside of other files to avoid circular imports."""

from abc import ABC, abstractmethod
from typing import Callable, Generic, Pattern, Optional, List, Tuple, TypeVar

from open_gopro.services import AttributeTable

NotiHandlerType = Callable[[int, bytes], None]

BleDevice = TypeVar("BleDevice")
BleClient = TypeVar("BleClient")


class BLEController(ABC, Generic[BleClient, BleDevice]):
    """Interface definition for a BLE driver to be used by GoPro."""

    @abstractmethod
    async def read(self, client: BleClient, uuid: str) -> bytearray:
        """Read a bytestream response from a UUID.

        Args:
            client (Client): client to pair to
            uuid (UUID): UUID to read from

        Returns:
            bytearray: response
        """
        raise NotImplementedError

    @abstractmethod
    async def write(self, client: BleClient, uuid: str, data: bytearray) -> None:
        """Write a bytestream to a UUID.

        Args:
            client (Client): client to pair to
            uuid (UUID): UUID to write to
            data (bytearray): bytestream to write

        Returns:
            bytearray: response
        """
        raise NotImplementedError

    @abstractmethod
    async def scan(self, token: Pattern, timeout: int = 5) -> BleDevice:
        """Scan BLE device with a regex in it's device name.

        Args:
            token (Pattern): Regex to scan for
            timeout (int, optional): [description]. Defaults to 5.

        Returns:
            Device: discovered device (shall not be multiple devices)
        """
        raise NotImplementedError

    @abstractmethod
    async def connect(self, device: BleDevice, timeout: int = 15) -> BleClient:
        """Connect to a BLE device.

        Args:
            device (Device): device to connect to
            timeout (int, optional): Timeout before considering connection establishment a failure. Defaults to 15.

        Returns:
            Client: client that has been connected to
        """
        raise NotImplementedError

    @abstractmethod
    async def pair(self, client: BleClient) -> None:
        """Pair to an already connected client.

        Args:
            client (Client): client to pair to
        """
        raise NotImplementedError

    @abstractmethod
    async def enable_notifications(self, client: BleClient, handler: NotiHandlerType) -> None:
        """Enable notifications for all notifiable characteristics.

        The handler is used to register for notifications. It will be called when a a notification
        is received.

        Args:
            client (Client): client to enable notifications on
            handler (Callable): notification handler
        """
        raise NotImplementedError

    @abstractmethod
    async def discover_chars(self, client: BleClient) -> AttributeTable:
        """Discover all characteristics for a connected client.

        Args:
            client (Client): client to discover on

        Returns:
            AttributeTable: dictionary of discovered services and characteristics indexed by UUID
        """
        raise NotImplementedError

    @abstractmethod
    async def disconnect(self, client: BleClient) -> bool:
        """Terminate the BLE connection.

        Args:
            client (Client): client to disconnect from
        """
        raise NotImplementedError


class WifiController(ABC):
    """Interface definition for a Wifi driver to be used by GoPro."""

    @abstractmethod
    def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """Connect to a network.

        Args:
            ssid (str): SSID of network to connect to
            password (str): password of network to connect to
            timeout (float, optional): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: True if successful, False otherwise
        """
        raise NotImplementedError

    @abstractmethod
    def disconnect(self) -> bool:
        """Disconnect from a network.

        Returns:
            bool: True if successful, False otherwise
        """
        raise NotImplementedError

    @abstractmethod
    def current(self) -> Tuple[Optional[str], Optional[str]]:
        """Return the SSID and state of the current network.

        Returns:
            Tuple[Optional[str], Optional[str]]: Tuple of SSID and state. State is optional. If SSID is None,
            there is no current connection
        """
        raise NotImplementedError

    @abstractmethod
    def interfaces(self) -> List[str]:
        """Return a list of wireless adapters.

        Returns:
            List[str]: adapters
        """
        raise NotImplementedError

    @abstractmethod
    def interface(self, interface: Optional[str]) -> Optional[str]:
        """Get or set the currently used wireless adapter.

        Args:
            interface (str, optional): Get if the interface parameter is None. Set otherwise. Defaults to None.

        Returns:
            Optional[str]: Name of interface if get. None if set.
        """
        raise NotImplementedError

    @abstractmethod
    def power(self, power: bool) -> None:
        """Enable / disable the wireless driver.

        Args:
            power (bool, optional): Enable if True. Disable if False.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def is_on(self) -> bool:
        """Is the wireless driver currently enabled.

        Returns:
            bool: True if yes. False if no.
        """
        raise NotImplementedError


class GoProError(Exception):
    """Base class for other exceptions."""

    def __init__(self, message: str) -> None:
        super().__init__(f"GoPro Error: {message}")


class ScanFailedToFindDevice(GoProError):
    """The scan failed without finding a device."""

    def __init__(self) -> None:
        super().__init__("A scan timed out without finding a device")


class ConnectFailed(GoProError):
    """A BLE or WiFi connection failed to establish

    Args:
        connection (str): type of connection that failed
        retries (int): how many retries were attempted
        timeout (int): the timeout used for each attempt
    """

    def __init__(self, connection: str, timeout: float, retries: int):
        super().__init__(
            f"{connection} connection failed to establish after {retries} retries with timeout {timeout}"
        )


class ResponseTimeout(GoProError):
    """A response has timed out."""

    def __init__(self, timeout: float) -> None:
        super().__init__(f"Response timeout occurred of {timeout} seconds")


class GoProNotInitialized(GoProError):
    """A command was attempted without waiting for the GoPro instance to initialize."""

    def __init__(self) -> None:
        super().__init__("GoPro has not been initialized yet")



# stream.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Entrypoint for taking a picture."""

import sys
import time
import logging
import argparse
import threading
from pathlib import Path
from typing import Tuple

from rich import traceback
from rich.logging import RichHandler
from rich.console import Console

from open_gopro import GoPro, params
from open_gopro.util import launch_vlc

logger = logging.getLogger(__name__)
traceback.install()  # Enable exception tracebacks in rich logger
console = Console()  # rich consoler printer


def main() -> None:
    """Main function."""
    identifier, log_location, vlc_location = parse_arguments()
    setup_logging(log_location)

    try:
        spinner = console.status("[bold green]Connecting...")
        spinner.start()
        with GoPro(identifier) as gopro:
            assert gopro.ble_command.set_shutter(params.Shutter.OFF).is_ok
            assert gopro.ble_command.set_turbo_mode(False).is_ok
            spinner.stop()

            with console.status("[bold green]Starting the preview stream..."):
                assert gopro.wifi_command.stop_preview_stream().is_ok
                assert gopro.wifi_command.start_preview_stream().is_ok

            with console.status("[bold green]Launching VLC..."):
                threading.Thread(target=launch_vlc, args=(vlc_location,), daemon=True).start()

            console.print(
                "Success!! :smiley: Stream has been enabled. VLC is viewing it at udp://@:8554",
                style="bold green",
            )
            console.print("Send keyboard interrupt to exit.")

            while True:
                time.sleep(0.2)

    except Exception as e:  # pylint: disable=broad-except
        logger.error(repr(e))
        sys.exit(-1)
    except KeyboardInterrupt:
        console.print("Received keyboard interrupt. Shutting down...")
    finally:
        console.print("Exiting...")
        sys.exit(0)


def setup_logging(log_location: Path) -> None:
    """Configure logging to file and Rich console logging

    Args:
        log_location (Path): location to configure for the file handler
    """
    # Logging to file with
    fh = logging.FileHandler(f"{log_location}", mode="w")
    file_formatter = logging.Formatter(
        fmt="%(threadName)13s: %(name)30s:%(lineno)5d %(asctime)s.%(msecs)03d %(levelname)-8s | %(message)s",
        datefmt="%H:%M:%S",
    )
    fh.setFormatter(file_formatter)
    fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)

    # Use Rich for colorful console logging
    sh = RichHandler(rich_tracebacks=True, enable_link_path=True, show_time=False)
    stream_formatter = logging.Formatter("%(asctime)s.%(msecs)03d %(message)s", datefmt="%H:%M:%S")
    sh.setFormatter(stream_formatter)
    sh.setLevel(logging.WARNING)
    logger.addHandler(sh)
    logger.setLevel(logging.DEBUG)

    # Enable / disable logging in modules
    for (module, level) in [
        ("open_gopro.gopro", logging.DEBUG),
        ("open_gopro.ble_commands", logging.DEBUG),
        ("open_gopro.ble_controller", logging.DEBUG),
        ("open_gopro.wifi_commands", logging.DEBUG),
        ("open_gopro.wifi_controller", logging.DEBUG),
        ("open_gopro.responses", logging.DEBUG),
        ("open_gopro.util", logging.DEBUG),
        ("bleak", logging.DEBUG),
        ("bleak.backends.bluezdbus.client", logging.DEBUG),
        ("bleak.backends.corebluetooth.client", logging.DEBUG),
        ("bleak.backends.dotnet.client", logging.DEBUG),
    ]:
        log = logging.getLogger(module)
        log.setLevel(level)
        log.addHandler(fh)
        log.addHandler(sh)


def parse_arguments() -> Tuple[str, Path, Path]:
    """Parse command line arguments

    Returns:
        Tuple[str, Path, Path]: (identifier, path to save log, path to VLC)
    """
    parser = argparse.ArgumentParser(
        description="Connect to a GoPro camera, enable the preview stream, then open VLC to view it."
    )
    parser.add_argument(
        "-i",
        "--identifier",
        type=str,
        help="Last 4 digits of GoPro serial number, which is the last 4 digits of the default camera SSID. If \
            not used, first discovered GoPro will be connected to",
        default=None,
    )
    parser.add_argument(
        "-l",
        "--log",
        type=Path,
        help="Location to store detailed log",
        default=Path("stream.log"),
    )
    parser.add_argument(
        "-v",
        "--vlc",
        type=Path,
        help="VLC location. If not set, the location will attempt to be automatically discovered.",
        default=None,
    )
    args = parser.parse_args()

    return args.identifier, args.log, args.vlc


if __name__ == "__main__":
    main()



# wifi_controller.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Manage a WiFI connection using native OS commands."""

# TODO This file needs to be cleaned up.

import os
import re
import time
import logging
import tempfile
from enum import Enum, auto
from typing import List, Optional, Tuple, Any

from open_gopro.util import cmd
from open_gopro.interfaces import WifiController

logger = logging.getLogger(__name__)


def cmp(a: Any, b: Any) -> int:
    """Define this since it is not implemented in Python 3.

    Args:
        a (Any): compare to
        b (Any): compare against

    Returns:
        int: compare result
    """
    if a < b:
        return -1

    if a == b:
        return 0

    return 1


def ensure_sudo() -> None:
    """Verify that we are running as root

    Raises:
        Exception: Program is running as a user other than root
    """
    user = cmd("whoami")
    if "root" not in user:
        logger.error(f"You need to be the root user to run this program but you are running as {user}")
        raise Exception("Program needs to be run as root user.")


class Wireless(WifiController):
    """Top level abstraction of different WiFi drivers.

    If interface is not specified (i.e. it is None), we will attempt ot automatically Receive response -->
    disover a suitable interface

    Args:
        interface (str, optional): Interface. Defaults to None.

    Raises:
        Exception: Failed to find a suitable driver or auto-detect the network interface.
    """

    _driver_name = "NOT INITIALIZED"
    _driver: WifiController

    # init
    def __init__(self, interface: Optional[str] = None):
        # detect and init appropriate driver
        self._driver_name = self._detectDriver()
        if self._driver_name == "nmcli":
            self._driver = NmcliWireless(interface=interface)
        elif self._driver_name == "nmcli0990":
            self._driver = Nmcli0990Wireless(interface=interface)
        elif self._driver_name == "wpa_supplicant":
            self._driver = WpasupplicantWireless(interface=interface)
        elif self._driver_name == "networksetup":
            self._driver = NetworksetupWireless(interface=interface)
        elif self._driver_name == "netsh":
            self._driver = NetshWireless(interface=interface)

        # attempt to auto detect the interface if none was provided
        if self.interface() is None:
            interfaces = self.interfaces()
            if len(interfaces) > 0:
                self.interface(interfaces[0])

        # raise an error if there is still no interface defined
        if self.interface() is None:
            raise Exception("Unable to auto-detect the network interface.")

        logger.debug(f"Using WiFi driver: {self._driver_name} with interface {self.interface()}")

    def _detectDriver(self) -> str:
        """Try to find a Wifi driver that can be used.

        Raises:
            Exception: We weren't able to find a suitable driver

        Returns:
            str: Name of discovered driver
        """
        if os.name == 'nt':
            # try netsh (Windows).
            response = cmd("where netsh")
            if len(response) > 0 and "not found" not in response and "not recognized" not in response:
                return "netsh"
            response = cmd("powershell get-command netsh")
            if len(response) > 0 and "not found" not in response and "not recognized" not in response:
                return "netsh"
        else:
            # try nmcli (Ubuntu 14.04)
            response = cmd("which nmcli")
            if len(response) > 0 and "not found" not in response:
                response = cmd("nmcli --version")
                parts = response.split()
                ver = parts[-1]
                compare = self.vercmp(ver, "0.9.9.0")
                if compare >= 0:
                    return "nmcli0990"

                return "nmcli"

            # try nmcli (Ubuntu w/o network-manager)
            response = cmd("which wpa_supplicant")
            if len(response) > 0 and "not found" not in response:
                return "wpa_supplicant"

        # try networksetup (Mac OS 10.10)
        response = cmd("which networksetup")
        if len(response) > 0 and "not found" not in response:
            return "networksetup"

        raise Exception("Unable to find compatible wireless driver.")

    @staticmethod
    def vercmp(actual: Any, test: Any) -> int:
        """Compare two versions.

        Args:
            actual (str): Version being compared
            test (str): Thing that version is being compared to

        Returns:
            -1: a is less than b
            0: a is equal to b
            1: a is greater than b
        """

        def normalize(v: str) -> List[int]:
            """Normalize a string vresion

            Args:
                v (str): input string

            Returns:
                List[int]: output int list
            """
            return [int(x) for x in re.sub(r"(\.0+)*$", "", v).split(".")]

        return cmp(normalize(actual), normalize(test))

    def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """Wrapper to call the OS-specific driver method.

        Args:
            ssid (str): network SSID
            password (str): network password
            timeout (float, optional): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: True if the connect was successful, False otherwise
        """
        return self._driver.connect(ssid, password)

    def disconnect(self) -> bool:
        """Wrapper to call the OS-specific driver method.

        Returns:
            bool: True if the disconnect was successful, False otherwise
        """
        return self._driver.disconnect()

    def current(self) -> Tuple[Optional[str], Optional[str]]:
        """Wrapper to call the OS-specific driver method.

        Returns:
            Tuple[Optional[str], Optional[str]]: (ssid, network state)
        """
        return self._driver.current()

    def interfaces(self) -> List[str]:
        """Wrapper to call the OS-specific driver method.

        Returns:
            List[str]: list of discovered interfaces
        """
        return self._driver.interfaces()

    def interface(self, interface: Optional[str] = None) -> Optional[str]:
        """Wrapper to call the OS-specific driver method.

        Use a str as interface to set it, otherwise use None to get it.

        Args:
            interface (Optional[str], optional): get or set interface. Defaults to None.

        Returns:
            Optional[str]: Str if getting
        """
        return self._driver.interface(interface)

    @property
    def is_on(self) -> bool:
        """Wrapper to call the OS-specific driver method.

        Returns:
            bool: True if on, False if off
        """
        return self._driver.is_on

    def power(self, power: bool) -> None:
        """Wrapper to call the OS-specific driver method.

        Args:
            power (bool): [description]

        Returns:
            [type]: [description]
        """
        return self._driver.power(power)

    def driver(self) -> str:
        """Get the name of the driver currently being used.

        Returns:
            str: Driver name.
        """
        return self._driver_name


class NmcliWireless(WifiController):
    """Linux nmcli Driver < 0.9.9.0."""

    _interface = None

    def __init__(self, interface: str = None) -> None:
        ensure_sudo()
        self.interface(interface)

    @staticmethod
    def _clean(partial: str) -> None:
        """Clean up connections.

        This is needed to prevent the following error after extended use:
        'maximum number of pending replies per connection has been reached'

        Args:
            partial (str): part of the connection name
        """
        # list matching connections
        response = cmd(f"nmcli --fields UUID,NAME con list | grep {partial}")

        # delete all of the matching connections
        for line in response.splitlines():
            if len(line) > 0:
                cmd(f"nmcli con delete uuid {line.split()[0]}")

    @staticmethod
    def _errorInResponse(response: str) -> bool:
        """Ignore warnings in nmcli output.

        Sometimes there are warnings but we connected just fine

        Args:
            response (str): output to parse

        Returns:
            bool: True if errors found. False if not.
        """
        # no error if no response
        if len(response) == 0:
            return False

        # loop through each line
        for line in response.splitlines():
            # all error lines start with 'Error'
            if line.startswith("Error"):
                return True

        # if we didn't find an error then we are in the clear
        return False

    def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """[summary].

        Args:
            ssid (str): network SSID
            password (str): network password
            timeout (float, optional): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: [description]
        """
        # clean up previous connection
        current, _ = self.current()
        if current is not None:
            self._clean(current)

        # attempt to connect
        response = cmd(
            "nmcli dev wifi connect {} password {} iface {}".format(ssid, password, self._interface)
        )

        # parse response
        return not self._errorInResponse(response)

    def disconnect(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return False

    def current(self) -> Tuple[Optional[str], Optional[str]]:
        """[summary].

        Returns:
            Tuple[Optional[str], Optional[str]]: [description]
        """
        # list active connections for all interfaces
        response = cmd("nmcli con status | grep {}".format(self.interface()))

        # the current network is in the first column
        for line in response.splitlines():
            if len(line) > 0:
                return (line.split()[0], None)

        # return none if there was not an active connection
        return (None, None)

    def interfaces(self) -> List[str]:
        """[summary].

        Returns:
            List[str]: [description]
        """
        # grab list of interfaces
        response = cmd("nmcli dev")

        # parse response
        interfaces = []
        for line in response.splitlines():
            if "wireless" in line:
                # this line has our interface name in the first column
                interfaces.append(line.split()[0])

        # return list
        return interfaces

    def interface(self, interface: Optional[str] = None) -> Optional[str]:
        """[summary].

        Args:
            interface (Optional[str], optional): [description]. Defaults to None.

        Returns:
            Optional[str]: [description]
        """
        if interface is not None:
            self._interface = interface
            return None
        return self._interface

    @property
    def is_on(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return "enabled" in cmd("nmcli nm wifi")

    def power(self, power: bool) -> None:
        """[summary].

        Args:
            power (bool): [description]
        """
        if power:
            cmd("nmcli nm wifi on")
        else:
            cmd("nmcli nm wifi off")


class Nmcli0990Wireless(WifiController):
    """Linux nmcli Driver >= 0.9.9.0."""

    _interface = None

    def __init__(self, interface: str = None):
        ensure_sudo()
        self.interface(interface)

    # TODO Is this needed?
    @staticmethod
    def _clean(partial: str) -> None:
        """Clean up connections.

        This is needed to prevent the following error after extended use:
        'maximum number of pending replies per connection has been reached'

        Args:
            partial (str): part of the connection name
        """
        # list matching connections
        response = cmd("nmcli --fields UUID,NAME con show | grep {}".format(partial))

        # delete all of the matching connections
        for line in response.splitlines():
            if len(line) > 0:
                uuid = line.split()[0]
                cmd("nmcli con delete uuid {}".format(uuid))

    @staticmethod
    def _errorInResponse(response: str) -> bool:
        """Ignore warnings in nmcli output.

        Sometimes there are warnings but we connected just fine

        Args:
            response (str): output to parse

        Returns:
            bool: True if errors found. False if not.
        """
        # no error if no response
        if len(response) == 0:
            return False

        # loop through each line
        for line in response.splitlines():
            # all error lines start with 'Error'
            if line.startswith("Error"):
                return True

        # if we didn't find an error then we are in the clear
        return False

    def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """[summary].

        Args:
            ssid (str): network SSID
            password (str): network password
            timeout (float, optional): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: [description]
        """
        # Scan for networks. Don't bother checking: we'll allow the error to be passed from the connect.
        cmd("nmcli dev wifi list --rescan yes")
        # attempt to connect
        response = cmd(f"nmcli dev wifi connect {ssid} password {password} ifname {self._interface}")

        # TODO verify that we're connected (and use timeout)

        # parse response
        return not self._errorInResponse(response)

    def disconnect(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return False

    def current(self) -> Tuple[Optional[str], Optional[str]]:
        """[summary].

        Returns:
            Tuple[Optional[str], Optional[str]]: [description]
        """
        # list active connections for all interfaces
        response = cmd("nmcli con | grep {}".format(self.interface()))

        # the current network is in the first column
        for line in response.splitlines():
            if len(line) > 0:
                return (line.split()[0], None)

        # return none if there was not an active connection
        return (None, None)

    def interfaces(self) -> List[str]:
        """[summary].

        Returns:
            List[str]: [description]
        """
        # grab list of interfaces
        response = cmd("nmcli dev")

        # parse response
        interfaces = []
        for line in response.splitlines():
            if "wifi" in line:
                # this line has our interface name in the first column
                interfaces.append(line.split()[0])

        # return list
        return interfaces

    def interface(self, interface: Optional[str] = None) -> Optional[str]:
        """[summary].

        Args:
            interface (Optional[str], optional): [description]. Defaults to None.

        Returns:
            Optional[str]: [description]
        """
        if interface is not None:
            self._interface = interface
            return None
        return self._interface

    @property
    def is_on(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return "enabled" in cmd("nmcli r wifi")

    def power(self, power: bool) -> None:
        """[summary].

        Args:
            power (bool): [description]
        """
        if power:
            cmd("nmcli r wifi on")
        else:
            cmd("nmcli r wifi off")


class WpasupplicantWireless(WifiController):
    """Linux wpa_supplicant Driver."""

    _file = "/tmp/wpa_supplicant.conf"
    _interface = None

    def __init__(self, interface: str = None):
        self.interface(interface)

    def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """[summary].

        Args:
            ssid (str): network SSID
            password (str): network password
            timeout (float, optional): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: [description]
        """
        # attempt to stop any active wpa_supplicant instances
        # ideally we do this just for the interface we care about
        cmd("sudo killall wpa_supplicant")

        # don't do DHCP for GoPros; can cause dropouts with the server
        cmd("sudo ifconfig {} 10.5.5.10/24 up".format(self._interface))

        # create configuration file
        f = open(self._file, "w")
        f.write('network={{\n    ssid="{}"\n    psk="{}"\n}}\n'.format(ssid, password))
        f.close()

        # attempt to connect
        cmd("sudo wpa_supplicant -i{} -c{} -B".format(self._interface, self._file))

        # check that the connection was successful
        # i've never seen it take more than 3 seconds for the link to establish
        time.sleep(5)
        current_ssid, _ = self.current()
        if current_ssid != ssid:
            return False

        # attempt to grab an IP
        # better hope we are connected because the timeout here is really long
        # cmd('sudo dhclient {}'.format(self._interface))

        # parse response
        return True

    def disconnect(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return False

    def current(self) -> Tuple[Optional[str], Optional[str]]:
        """[summary].

        Returns:
            Tuple[Optional[str], Optional[str]]: [description]
        """
        # get interface status
        response = cmd("iwconfig {}".format(self.interface()))

        # the current network is on the first line like ESSID:"network"
        line = response.splitlines()[0]
        line = line.replace('"', "")
        parts = line.split("ESSID:")
        if len(parts) > 1:
            network = parts[1].strip()
            if network != "off/any":
                return (network, None)

        # return none if there was not an active connection
        return (None, None)

    def interfaces(self) -> List[str]:
        """[summary].

        Returns:
            List[str]: [description]
        """
        # grab list of interfaces
        response = cmd("iwconfig")

        # parse response
        interfaces = []
        for line in response.splitlines():
            if len(line) > 0 and not line.startswith(" "):
                # this line contains an interface name!
                if "no wireless extensions" not in line:
                    # this is a wireless interface
                    interfaces.append(line.split()[0])
        return interfaces

    def interface(self, interface: Optional[str] = None) -> Optional[str]:
        """[summary].

        Args:
            interface (Optional[str], optional): [description]. Defaults to None.

        Returns:
            Optional[str]: [description]
        """
        if interface is not None:
            self._interface = interface
            return None
        return self._interface

    @property
    def is_on(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        # TODO
        return True

    def power(self, power: bool) -> None:
        """[summary].

        Args:
            power (bool): [description]
        """
        # TODO
        return


class NetworksetupWireless(WifiController):
    """OS X networksetup Driver."""

    _interface = None

    def __init__(self, interface: str = None):
        self.interface(interface)

    def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """[summary].

        Args:
            ssid (str): network SSID
            password (str): network password
            timeout (float, optional): Time before considering connection failed (in seconds). Defaults to 15.

        Raises:
            Exception: [description]

        Returns:
            bool: [description]
        """
        # Escape single quotes
        ssid = ssid.replace(r"'", '''"'"''')
        response = cmd(
            "networksetup -setairportnetwork '{}' '{}' '{}'".format(self._interface, ssid, password)
        )

        if "not find" in response.lower():
            return False
        # Now wait for network to actually establish
        current = self.current()[0]
        logger.debug(f"current wifi: {current}")
        while current is not None and ssid not in current and timeout < 10:
            time.sleep(1)
            current = self.current()[0]
            logger.debug(f"current wifi: {current}")
            timeout -= 1
            if timeout == 0:
                raise Exception("Wi-Fi connection timeout.")

        # TODO There is some delay required here, presumably because the network is not ready.
        time.sleep(5)

        return True

    def disconnect(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return False

    def current(self) -> Tuple[Optional[str], Optional[str]]:
        """[summary].

        Returns:
            Tuple[Optional[str], Optional[str]]: [description]
        """
        # attempt to get current network
        response = cmd("networksetup -getairportnetwork {}".format(self._interface))

        # parse response
        phrase = "Current Wi-Fi Network: "
        if phrase in response:
            return (response.replace("Current Wi-Fi Network: ", "").strip(), None)
        return (None, None)

    def interfaces(self) -> List[str]:
        """[summary].

        Returns:
            List[str]: [description]
        """
        # grab list of interfaces
        response = cmd("networksetup -listallhardwareports")

        # parse response
        interfaces = []
        detectedWifi = False
        for line in response.splitlines():
            if detectedWifi:
                # this line has our interface name in it
                interfaces.append(line.replace("Device: ", ""))
                detectedWifi = False
            else:
                # search for the line that has 'Wi-Fi' in it
                if "Wi-Fi" in line:
                    detectedWifi = True

        # return list
        return interfaces

    def interface(self, interface: Optional[str] = None) -> Optional[str]:
        """[summary].

        Args:
            interface (Optional[str], optional): [description]. Defaults to None.

        Returns:
            Optional[str]: [description]
        """
        if interface is not None:
            self._interface = interface
            return None
        return self._interface

    @property
    def is_on(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return "On" in cmd("networksetup -getairportpower {}".format(self._interface))

    def power(self, power: bool) -> None:
        """[summary].

        Args:
            power (bool): [description]
        """
        if power:
            cmd("networksetup -setairportpower {} on".format(self._interface))
        else:
            cmd("networksetup -setairportpower {} off".format(self._interface))


class NetshWireless(WifiController):
    """Windows Driver."""

    # Used to build profile
    template = r"""<?xml version="1.0"?>
<WLANProfile xmlns="http://www.microsoft.com/networking/WLAN/profile/v1">
    <name>{ssid}</name>
    <SSIDConfig>
        <SSID>
            <name>{ssid}</name>
        </SSID>
    </SSIDConfig>
    <connectionType>ESS</connectionType>
    <connectionMode>manual</connectionMode>
    <MSM>
        <security>
            <authEncryption>
                <authentication>{auth}</authentication>
                <encryption>{encrypt}</encryption>
                <useOneX>false</useOneX>
            </authEncryption>
            <sharedKey>
                <keyType>passPhrase</keyType>
                <protected>false</protected>
                <keyMaterial>{passwd}</keyMaterial>
            </sharedKey>
        </security>
    </MSM>
    <MacRandomization xmlns="http://www.microsoft.com/networking/WLAN/profile/v3">
        <enableRandomization>false</enableRandomization>
    </MacRandomization>
</WLANProfile>"""

    def __init__(self, interface: str = None) -> None:
        self._interface: Optional[str] = None
        self.interface(interface)
        self.ssid: Optional[str] = None

    def __del__(self) -> None:
        # TODO Do we want this?
        # self._clean(self.ssid)
        pass

    def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """Establish a connection.

        This is blocking and won't return until either a connection is established or
        a 10 second timeout

        Args:
            ssid (str): SSID of network to connect to
            password (str): password of network to connect to
            timeout (float, optional): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: True if connected, False otherwise
        """
        # Replace ampersand as it causes problems
        password = password.replace("&", "&amp;")

        logger.info(f"Attempting to establish WiFi connection to {ssid}...")

        # Start fresh each time.
        self._clean(ssid)

        # Create new profile
        output = NetshWireless.template.format(ssid=ssid, auth="WPA2PSK", encrypt="AES", passwd=password)
        logger.debug(output)
        # Need ugly low level mkstemp and os here because standard tempfile can't be accessed by a subprocess in Windows :(
        fd, filename = tempfile.mkstemp()
        os.write(fd, output.encode("utf-8"))
        os.close(fd)
        response = cmd(f"netsh wlan add profile filename={filename}")
        if "is added on interface" not in response:
            raise Exception(response)
        os.remove(filename)

        # Try to connect
        ssid_quotes = f'"{ssid}"'
        response = cmd(f"netsh wlan connect ssid={ssid_quotes} name={ssid_quotes} interface={self._interface}")
        if "was completed successfully" not in response:
            raise Exception(response)

        while self.current() != (ssid, "connected"):
            logger.debug("Waiting 1 second for Wi-Fi connection to establish...")
            time.sleep(1)
            timeout -= 1
            if timeout == 0:
                raise Exception("Wi-Fi connection timeout.")

        logger.info("Wifi connection established!")
        self.ssid = ssid

        return True

    def disconnect(self) -> bool:
        """Terminate the WiFi connection.

        Returns:
            bool: True if the disconnect was successful, False otherwise.
        """
        response = cmd(f"netsh wlan disconnect interface={self.interface()}")

        return bool("completed successfully" in response.lower())

    def current(self) -> Tuple[Optional[str], Optional[str]]:
        """Get the current network SSID and state.

        # Here is an example of what we are parsing (i.e. to find FunHouse SSID):
        # Name                   : Wi-Fi
        # Description            : TP-Link Wireless USB Adapter
        # GUID                   : 093d8022-33cb-4400-8362-275eaf24cb86
        # Physical address       : 98:48:27:88:cb:18
        # State                  : connected
        # SSID                   : FunHouse

        Raises:
            Exception: Unexpected error.

        Returns:
            Tuple[Optional[str], Optional[str]]: Tuple of (ssid, network_state)
        """

        class ParseState(Enum):
            """Current state of interface parsing"""

            PARSE_INTERFACE = auto()
            PARSE_SSID = auto()
            PARSE_STATE = auto()

        if self.interface is None:
            self._interface = self.interfaces()[0]
        if self._interface is None:
            raise Exception("Can't auto-assign interface. None found.")

        response = cmd("netsh wlan show interfaces")
        parse_state = ParseState.PARSE_INTERFACE
        ssid: Optional[str] = None
        network_state: Optional[str] = None
        for field in response.split("\r\n"):
            if parse_state is ParseState.PARSE_INTERFACE:
                if "Name" in field and self._interface in field:
                    parse_state = ParseState.PARSE_STATE
            elif parse_state is ParseState.PARSE_STATE:
                if "State" in field:
                    network_state = field.split(":")[1].strip()
                    parse_state = ParseState.PARSE_SSID
            elif parse_state is ParseState.PARSE_SSID:
                if "SSID" in field:
                    ssid = field.split(":")[1].strip()
                    break

        return (ssid, network_state)

    def interfaces(self) -> List[str]:
        """Discover all available interfaces.

        # We're parsing, for example, the following line to find "Wi-Fi":
        # Name                   : Wi-Fi

        Returns:
            List[str]: List of interfaces
        """
        response = cmd("netsh wlan show interfaces")
        interfaces = []

        # Look behind to find field, then match (non-greedy) any chars until CRLF
        match = "(?<={}).+?(?=\\r\\n)"
        for interface in re.findall(match.format("Name"), response):
            # Strip leading whitespace and then the first two chars of remaining (i.e. " :")
            interfaces.append(interface.strip()[2:])

        return interfaces

    def interface(self, interface: str = None) -> Optional[str]:
        """Get or set the current interface.

        Args:
            interface (str, optional): String to set or None to get. Defaults to None.

        Returns:
            Optional[str]: If interface argument is None, this will be a string if there is a valid interface; otherwise None
        """
        if interface is not None:
            self._interface = interface
            return None
        return self._interface

    @property
    def is_on(self) -> bool:
        """Is Wifi enabled?

        Returns:
            bool: True if yes, False if no.
        """
        # TODO
        return True

    def power(self, power: bool) -> None:
        """Enable / Disable WiFi.

        Args:
            power (bool): True to enable, False to disable.
        """
        arg = "enable" if power is True else "disable"
        cmd(f"netsh interface set interface {self._interface} {arg}")

    @staticmethod
    def _clean(ssid: Optional[str]) -> None:
        """Disconnect and delete SSID profile.

        Args:
            ssid (Optional[str]): name of SSID
        """
        cmd("netsh wlan disconnect")
        if ssid is not None:
            cmd(f'netsh wlan delete profile name="{ssid}"')



# params.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Parameter definitions for GoPro BLE and WiFi commands.

Note these have to be IntEnum's in order to be correctly build when sending commands
"""

import enum


class Shutter(enum.IntEnum):  # pylint: disable=missing-class-docstring
    ON = 1
    OFF = 0


class Preset(enum.IntEnum):  # pylint: disable=missing-class-docstring
    ACTIVITY = 1
    BURST_PHOTO = 65538
    CINEMATIC = 2
    LIVE_BURST = 65537
    NIGHT_PHOTO = 65539
    NIGHT_LAPSE = 131074
    PHOTO = 65536
    SLO_MO = 3
    STANDARD = 0
    TIME_LAPSE = 131073
    TIME_WARP = 131072
    MAX_PHOTO = 262144
    MAX_TIMEWARP = 327680
    MAX_VIDEO = 196608


class PresetGroup(enum.IntEnum):  # pylint: disable=missing-class-docstring
    VIDE0 = 1000
    PHOTO = 1001
    TIMELAPSE = 1002


class Resolution(enum.IntEnum):  # pylint: disable=missing-class-docstring
    RES_4k = 1
    RES_2_7k = 4
    RES_2_7k_4_3 = 6
    RES_1440 = 7
    RES_1080 = 9
    RES_4k_4_3 = 18
    RES_5_6k = 15
    RES_3k = 16
    RES_5k = 24


class FPS(enum.IntEnum):  # pylint: disable=missing-class-docstring
    FPS_240 = 0
    FPS_120 = 1
    FPS_100 = 2
    FPS_60 = 5
    FPS_50 = 6
    FPS_30 = 8
    FPS_25 = 9
    FPS_24 = 10
    FPS_200 = 13


class LensMode(enum.IntEnum):  # pylint: disable=missing-class-docstring
    SINGLE = 0
    DUAL = 5


class FieldOfView(enum.IntEnum):  # pylint: disable=missing-class-docstring
    MAX_SUPERVIEW = 7
    SUPERVIEW = 3
    WIDE = 0
    LINEAR = 4
    NARROW = 6
    LINEAR_HORIZON = 8


class LED(enum.IntEnum):  # pylint: disable=missing-class-docstring
    ALL_ON = 2
    ALL_OFF = 0
    FRONT_OFF = 1
    BLE_KEEP_ALIVE = 61


class PairState(enum.IntEnum):  # pylint: disable=missing-class-docstring
    SUCCESS = 0
    IN_PROGRESS = 1
    FAILED = 2
    STOPPED = 3


class PairType(enum.IntEnum):  # pylint: disable=missing-class-docstring
    NOT_PAIRING = 0
    PAIRING_APP = 1
    PAIRING_REMOTE_CONTROL = 2
    PAIRING_BLUETOOTH = 3


class WAPState(enum.IntEnum):  # pylint: disable=missing-class-docstring
    NEVER_STARTED = 0
    STARTED = 1
    ABORTED = 2
    CANCELED = 3
    COMPLETED = 4


class SDStatus(enum.IntEnum):  # pylint: disable=missing-class-docstring
    OK = 0
    FULL = 1
    REMOVED = 2
    FORMAT_ERROR = 3
    BUSY = 4
    SWAPPED = 8
    UNKNOWN = 0xFF


class OTAStatus(enum.IntEnum):  # pylint: disable=missing-class-docstring
    IDLE = 0
    DOWNLOADING = 1
    VERIFYING = 2
    DOWNLOAD_FAILED = 3
    VERIFY_FAILED = 4
    READY = 5
    GOPRO_APP_DOWNLOADING = 6
    GOPRO_APP_VERIFYING = 7
    GOPRO_APP_DOWNLOAD_FAILED = 8
    GOPRO_APP_VERIFY_FAILED = 9
    GOPRO_APP_READY = 10


class AnalyticsState(enum.IntEnum):  # pylint: disable=missing-class-docstring
    NOT_READY = 0
    READY = 1
    ON_CONNECT = 2


class ExposureMode(enum.IntEnum):  # pylint: disable=missing-class-docstring
    DISABLED = 0
    AUTO = 1
    ISO_LOCK = 2
    HEMISPHERE = 3


class AccMicStatus(enum.IntEnum):  # pylint: disable=missing-class-docstring
    NOT_CONNECTED = 0
    CONNECTED = 1
    CONNECTED_AND_PLUGGED_IN = 2


class WiFiBand(enum.IntEnum):  # pylint: disable=missing-class-docstring
    BAND_2_4_GHZ = 0
    BAND_5_GHZ = 1
    BAND_MAX = 2


class Orientation(enum.IntEnum):  # pylint: disable=missing-class-docstring
    UPRIGHT = 0
    UPSIDE_DOWN = 1
    ON_RIGHT_SIDE = 2
    ON_LEFT_SIDE = 3


class MediaModMicStatus(enum.IntEnum):  # pylint: disable=missing-class-docstring
    REMOVED = 0
    ONLY = 1
    WITH_EXTERNAL = 2


class TimeWarpSpeed(enum.IntEnum):  # pylint: disable=missing-class-docstring
    SPEED_15X = 0
    SPEED_30x = 1
    SPEED_60X = 2
    SPEED_150X = 3
    SPEED_300X = 4
    SPEED_900X = 5
    SPEED_1800X = 6
    SPEED_2X = 7
    SPEED_5X = 8
    SPEED_10X = 9
    SPEED_AUTO = 10
    SPEED_1X = 11
    SPEED_HALF = 12


class MaxLensMode(enum.IntEnum):  # pylint: disable=missing-class-docstring
    DEFAULT = 0
    MAX_LENS = 1


class MediaModStatus(enum.IntEnum):  # pylint: disable=missing-class-docstring
    SELFIE_0_HDMI_0_MEDIAMODCONNECTED_FALSE = 0
    SELFIE_0_HDMI_0_MEDIAMODCONNECTED_TRUE = 1
    SELFIE_0_HDMI_1_MEDIAMODCONNECTED_FALSE = 2
    SELFIE_0_HDMI_1_MEDIAMODCONNECTED_TRUE = 3
    SELFIE_1_HDMI_0_MEDIAMODCONNECTED_FALSE = 4
    SELFIE_1_HDMI_0_MEDIAMODCONNECTED_TRUE = 5
    SELFIE_1_HDMI_1_MEDIAMODCONNECTED_FALSE = 6
    SELFIE_1_HDMI_1_MEDIAMODCONNECTED_TRUE = 7


class Shortcut(enum.IntEnum):  # pylint: disable=missing-class-docstring
    HORIZON_LOCK = 24
    WIND = 23
    MEDIA_MOD = 16
    RAW_AUDIO = 18
    COLOR = 10
    SHARPNESS = 19
    ISO_MAX = 15
    ISO_MIN = 14
    WHITE_BALANCE = 5
    EV_COMP = 11
    SHUTTER = 6
    BIT_RATE = 7
    ZOOM = 1
    TIMER = 20
    BURST_RATE = 8
    HINDSIGHT = 27
    DURATION = 9
    SCHEDULE_CAPTURE = 26
    HYPERSMOOTH_BOOST = 12
    SPEED_RAMP = 25
    SPEED = 21
    OUTPUT = 17
    INTERVAL = 13
    LOOPING_INTERVAL = 13
    LENS = 2
    SLOMO = 3
    OFF = 0


class Flatmode(enum.IntEnum):  # pylint: disable=missing-class-docstring
    VIDEO = 12
    LOOPING = 15
    PHOTO = 17
    NIGHT_PHOTO = 18
    BURST_PHOTO = 19
    TIME_LAPSE_VIDEO = 13
    TIME_LAPSE_PHOTO = 20
    NIGHT_LAPSE_PHOTO = 21
    TIME_WARP_VIDEO = 24
    LIVE_BURST = 25
    NIGHT_LAPSE_VIDEO = 26
    SLO_MO = 27


class Toggle(enum.IntEnum):  # pylint: disable=missing-class-docstring
    ENABLE = 1
    DISABLE = 0



# photo.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Entrypoint for taking a picture."""

import sys
import logging
import argparse
from pathlib import Path
from typing import Tuple

from rich import traceback
from rich.logging import RichHandler
from rich.console import Console

from open_gopro import GoPro, params

logger = logging.getLogger(__name__)
traceback.install()  # Enable exception tracebacks in rich logger
console = Console()  # rich consoler printer


def main() -> None:
    """Main function."""
    identifier, log_location, output_location = parse_arguments()
    setup_logging(log_location)

    try:
        spinner = console.status("[bold green]Connecting...")
        spinner.start()
        with GoPro(identifier) as gopro:
            assert gopro.ble_command.set_shutter(params.Shutter.OFF).is_ok
            assert gopro.ble_command.set_turbo_mode(False).is_ok
            spinner.stop()

            with console.status("[bold green]Capturing a photo..."):
                # Get the media list before
                media_set_before = set(x["n"] for x in gopro.wifi_command.get_media_list()["media"][0]["fs"])
                # Take a photo
                assert gopro.ble_command.load_preset(params.Preset.PHOTO).is_ok
                assert gopro.ble_command.set_shutter(params.Shutter.ON).is_ok

            with console.status("[bold green]Downloading the photo..."):
                # Get the media list after
                media_set_after = set(x["n"] for x in gopro.wifi_command.get_media_list()["media"][0]["fs"])
                # The photo (is most likely) the difference between the two dicts as sets
                photo = media_set_after.difference(media_set_before).pop()
                # Download the photo
                gopro.wifi_command.download_file(camera_file=photo, local_file=output_location)

            console.print(
                f"Success!! :smiley: File has been downloaded to {output_location}", style="bold green"
            )

    except Exception as e:  # pylint: disable=broad-except
        logger.error(repr(e))
        sys.exit(-1)
    except KeyboardInterrupt:
        logger.warning("Received keyboard interrupt. Shutting down...")
    finally:
        console.print("Exiting...")
        sys.exit(0)


def setup_logging(log_location: Path) -> None:
    """Configure logging to file and Rich console logging

    Args:
        log_location (Path): location to configure for the file handler
    """
    # Logging to file with
    fh = logging.FileHandler(f"{log_location}", mode="w")
    file_formatter = logging.Formatter(
        fmt="%(threadName)13s: %(name)30s:%(lineno)5d %(asctime)s.%(msecs)03d %(levelname)-8s | %(message)s",
        datefmt="%H:%M:%S",
    )
    fh.setFormatter(file_formatter)
    fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)

    # Use Rich for colorful console logging
    sh = RichHandler(rich_tracebacks=True, enable_link_path=True, show_time=False)
    stream_formatter = logging.Formatter("%(asctime)s.%(msecs)03d %(message)s", datefmt="%H:%M:%S")
    sh.setFormatter(stream_formatter)
    sh.setLevel(logging.WARNING)
    logger.addHandler(sh)
    logger.setLevel(logging.DEBUG)

    # Enable / disable logging in modules
    for (module, level) in [
        ("open_gopro.gopro", logging.DEBUG),
        ("open_gopro.ble_commands", logging.DEBUG),
        ("open_gopro.ble_controller", logging.DEBUG),
        ("open_gopro.wifi_commands", logging.DEBUG),
        ("open_gopro.wifi_controller", logging.DEBUG),
        ("open_gopro.responses", logging.DEBUG),
        ("open_gopro.util", logging.DEBUG),
        ("bleak", logging.DEBUG),
        ("bleak.backends.bluezdbus.client", logging.DEBUG),
        ("bleak.backends.corebluetooth.client", logging.DEBUG),
        ("bleak.backends.dotnet.client", logging.DEBUG),
    ]:
        log = logging.getLogger(module)
        log.setLevel(level)
        log.addHandler(fh)
        log.addHandler(sh)


def parse_arguments() -> Tuple[str, Path, Path]:
    """Parse command line arguments

    Returns:
        Tuple[str, Path, Path]: (identifier, path to save log, path to store photo)
    """
    parser = argparse.ArgumentParser(description="Connect to a GoPro camera, take a photo, then download it.")
    parser.add_argument(
        "-i",
        "--identifier",
        type=str,
        help="Last 4 digits of GoPro serial number, which is the last 4 digits of the default camera SSID. If \
            not used, first discovered GoPro will be connected to",
        default=None,
    )
    parser.add_argument(
        "-l",
        "--log",
        type=Path,
        help="Location to store detailed log",
        default=Path("photo.log"),
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        help="Where to write the photo to. If not set, write to 'photo.jpg'",
        default=Path("photo.jpg"),
    )
    args = parser.parse_args()

    return args.identifier, args.log, args.output


if __name__ == "__main__":
    main()



# responses.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Any responses that are returned from GoPro commands."""

import enum
import json
import logging
from collections import defaultdict
from abc import abstractmethod, ABC
from typing import (
    Any,
    ClassVar,
    Dict,
    Callable,
    List,
    Optional,
    Union,
    Iterator,
    Type,
    ItemsView,
    ValuesView,
    KeysView,
)

import requests
import construct  # For enum
from construct import Struct, FormatField, StringEncoded

from open_gopro.constants import (
    ActionId,
    StatusId,
    UUID,
    CmdId,
    ErrorCode,
    SettingId,
    QueryCmdId,
    ResponseType,
    CmdType,
)
from open_gopro.util import scrub

logger = logging.getLogger(__name__)


# Type of data that will be parsed into a GoPro Response.
InputType = Optional[Union[bytearray, Dict[str, Any]]]


class Parser(ABC):
    """Interface definition for a non-Construct parser to be used by GoProResp for parsing."""

    @abstractmethod
    def parse(self, buf: Any) -> Dict[Any, Any]:
        """Parse any input type into a dict.

        Args:
            buf (InputType): input to parse

        Returns:
            Dict[Any, Any]: parsed output
        """
        raise NotImplementedError


class Builder(ABC):
    """Interface definition for a non-Construct builder to be used by GoProResp for building."""

    @abstractmethod
    def build(self, buf: Any) -> bytearray:
        """Build something into a bytestream.

        Args:
            buf (Any): The thing to build

        Returns:
            bytearray: The build bytestream.
        """
        raise NotImplementedError


class ParserBuilder(Parser, Builder):
    """A class that can both parse and build."""

    ...


FieldBuilder = Callable[..., FormatField]
ConstructType = Union[Struct, construct.Enum, FieldBuilder, StringEncoded]

ParserBuilderType = Union[ParserBuilder, ConstructType]
ParserType = Union[Parser, ParserBuilderType, ConstructType]
BuilderType = Union[Builder, ParserBuilderType, ConstructType]


CONT_MASK = 0b10000000
HDR_MASK = 0b01100000
GEN_LEN_MASK = 0b00011111
EXT_13_BYTE0_MASK = 0b00011111

id_map: Dict[UUID, Callable] = {
    UUID.CQ_SETTINGS_RESP: SettingId,
    UUID.CQ_COMMAND_RESP: CmdId,
    UUID.CQ_QUERY_RESP: QueryCmdId,
}

response_map: Dict[UUID, UUID] = {
    UUID.CQ_SETTINGS: UUID.CQ_SETTINGS_RESP,
    UUID.CQ_COMMAND: UUID.CQ_COMMAND_RESP,
    UUID.CQ_QUERY: UUID.CQ_QUERY_RESP,
}


class GoProResp:
    """A flexible object to be used to encapsulate all GoPro responses.

    It can be instantiated with varying levels of information and filled out as more is received.

    It is mostly a wrapper around a JSON-like dictionary (GoProResp.data)

    >>> response = ble_setting.resolution.get_value()
    >>> print(response.status)
    ErrorCode.SUCCESS
    >>> print(response.is_ok)
    True
    >>> print(response.id)
    QueryCmdId.GET_SETTING_VAL
    >>> print(response.cmd)
    QueryCmdId.GET_SETTING_VAL
    >>> print(response.uuid)
    UUID.CQ_QUERY_RESP
    >>> print(response.data)
    {
        "status": "SUCCESS",
        "id": "UUID.CQ_QUERY_RESP::QueryCmdId.GET_SETTING_VAL",
        "SettingId.RESOLUTION": [
            "RES_1080"
        ]
    }

    Args:
        info (List[ResponseType]): A list of all information known about the response.
        status (ErrorCode, optional): A status if known at time of instantiation. Defaults to ErrorCode.SUCCESS.
        raw_packet (InputType, optional): The unparsed input if known at time of instantiation. Defaults to None.
    """

    # To be filled out when settings / statuses are defined
    parser_map: ClassVar[Dict[ResponseType, ParserType]] = {}

    class State(enum.Enum):
        """Describes the state of the GoProResp."""

        INITIALIZED = enum.auto()
        ACCUMULATED = enum.auto()
        PARSED = enum.auto()

    def __init__(
        self,
        info: List[ResponseType],
        status: ErrorCode = ErrorCode.SUCCESS,
        raw_packet: InputType = None,
    ) -> None:
        self._info: List[ResponseType] = info
        """A list describing all of the currently known information about the response.

        This will be appended to as more information is discovered. The various properties of GoProResp will
        use this list to parse out their relevant information.
        """

        self.status: ErrorCode = status
        """Status of the response"""

        # Start with empty list as default value in case we need to append.
        # If we end up not needing a list, we will just overwrite the default
        self.data: Dict[Any, Any] = defaultdict(list)
        """Response data which is really JSON data stored as a dict"""

        self._raw_packet = raw_packet
        self._bytes_remaining = 0
        self._state = GoProResp.State.INITIALIZED

    @classmethod
    def from_write_command(cls, uuid: UUID, data: bytes) -> "GoProResp":
        """Build a GoProResp from a write command.

        This will discover the expected response UUID from the command UUID as well as parse the ID
        from the bytestream.

        Args:
            uuid (UUID): UUID that write command is writing to
            data (bytes): bytestream of the command

        Returns:
            GoProResp: created instance
        """
        # Find expected response and ID from the data
        response_uuid = response_map[uuid]
        cmd_id = id_map[response_uuid](data[1])
        info = [response_uuid, cmd_id]
        # If this is a protobuf command, it also has an action id
        if cmd_id.value >= 0xF0:
            info.append(ActionId(data[2]))
        return cls(info=info)

    @classmethod
    def from_read_response(cls, uuid: UUID, data: bytearray) -> "GoProResp":
        """Build a GoProResp from a read response.

        Args:
            uuid (UUID): UUID that read command was received from
            data (bytes): received bytestream

        Returns:
            GoProResp: created instance
        """
        resp = cls(info=[uuid], status=ErrorCode.SUCCESS, raw_packet=data)
        resp._parse()
        return resp

    @classmethod
    def from_http_response(cls, response: requests.models.Response) -> "GoProResp":
        """Build a GoProResp from an HTTP response from the requests package.

        Args:
            response (requests.models.Response): HTTP response

        Returns:
            GoProResp: created instance
        """
        resp = cls(
            info=[response.request.path_url.strip("/")],
            status=ErrorCode.SUCCESS if response.ok else ErrorCode.ERROR,
            raw_packet=response.json(),
        )
        resp._parse()
        return resp

    def __getitem__(self, key: Any) -> Any:  # pylint: disable=missing-return-doc
        return self.data[key]

    def __contains__(self, item: Any) -> bool:  # pylint: disable=missing-return-doc
        return item in self.data

    def __iter__(self) -> Iterator:  # pylint: disable=missing-return-doc
        return iter(self.data)

    def __str__(self) -> str:  # pylint: disable=missing-return-doc
        return json.dumps(
            {
                "status": self.status.name,
                "id": "::".join([str(x) for x in self._info]),
                **{str(k): v for k, v in self.data.items()},
            },
            default=str,
            indent=4,
        )

    def items(self) -> ItemsView[Any, Any]:
        """Access data dict's "items" method

        Returns:
            ItemsView[Any, Any]: [description]
        """
        return self.data.items()

    def keys(self) -> KeysView[Any]:
        """Access data dict's "keys" method

        Returns:
            ItemsView[Any, Any]: [description]
        """
        return self.data.keys()

    def values(self) -> ValuesView[Any]:
        """Access data dict's "values" method

        Returns:
            ItemsView[Any, Any]: [description]
        """
        return self.data.values()

    @property
    def is_received(self) -> bool:
        """Has the response been completely received?

        Returns:
            bool: True if completely received, False if not
        """
        return self._state is GoProResp.State.ACCUMULATED or self._state is GoProResp.State.PARSED

    @property
    def is_parsed(self) -> bool:
        """Has the response been successfully parsed?

        Returns:
            bool: True if it has been parsed, False if not
        """
        return self._state is GoProResp.State.PARSED

    @property
    def is_ok(self) -> bool:
        """Are there any errors in this response?

        Returns:
            bool: True if the response is ok (i.e. there are no errors), False otherwise
        """
        return self.status is ErrorCode.SUCCESS

    @property
    def id(self) -> ResponseType:
        """Get the identifier of the response.

        Will vary depending on what type of response this is:

        - for a direct BLE read / write to a characteristic, it will be a :py:class:`open_gopro.constants.UUID`
        - for a BLE command response, it will be a :py:class:`open_gopro.constants.CmdType`
        - for a BLE setting / status response / update, it will be a :py:class:`open_gopro.constants.SettingId` or :py:class:`open_gopro.constants.StatusId`
        - for an HTTP response, it will be a string of the HTTP endpoint

        Returns:
            ResponseType: the identifier
        """
        return self._info[-1]

    @property
    def cmd(self) -> Optional[CmdType]:
        """Attempt to get the command ID of the response.

        If the response is not a command response, it won't have a command.

        Returns:
            Optional[CmdType]: Command ID if relevant, otherwise None
        """
        for x in self._info:
            if isinstance(x, (QueryCmdId, CmdId)):
                return x  # type: ignore
        return None

    @property
    def uuid(self) -> Optional[UUID]:
        """Attempt to get the UUID the response was received on.

        If the response is not a BLE response, it won't have a UUID.

        Returns:
            Optional[UUID]: UUID if relevant, otherwise None
        """
        for x in self._info:
            if isinstance(x, UUID):
                return x
        return None

    @property
    def endpoint(self) -> Optional[str]:
        """Attempt to get the endpoint that response was received from.

        If the response is not an HTTP response, it won't have an endpoint.

        Returns:
            Optional[str]: Endpoint if relevant, otherwise None
        """
        for x in self._info:
            if isinstance(x, str):
                return x
        return None

    @property
    def flatten(self) -> Any:
        """Attempt to flatten / simplify the JSON response (GoProResp.data).

        - If there is only one entry in the JSON dict and it is a single value, return the value
        - If there is only one entry in the JSON dict and it is a list of values, return the list
        - Otherwise, just return the JSON dict

        Returns:
            Any: A single value, a list of values, or the JSON dict
        """
        values = list(self.data.values())
        if len(values) == 1 and type(values[0] not in [list, dict]):
            return values[0]

        if len(values) == 1 and type(values[0] is list):
            return values

        return self.data

    # to be @overload'ed if anything besides BLE notifications need to be accumulated
    def _accumulate(self, data: bytes) -> None:
        """Accumulate BLE byte data.

        If there is no more data left to accumulate, the bytestream will be parsed.

        Args:
            data (bytes): byte level BLE data
        """

        class Header(enum.Enum):
            """Packet Headers."""

            GENERAL = 0b00
            EXT_13 = 0b01
            EXT_16 = 0b10
            RESERVED = 0b11
            CONT = enum.auto()

        buf = bytearray(data)
        if buf[0] & CONT_MASK:
            buf.pop(0)
        else:
            # This is a new packet so start with an empty byte array
            self._raw_packet = bytearray([])
            hdr = Header((buf[0] & HDR_MASK) >> 5)
            if hdr is Header.GENERAL:
                self._bytes_remaining = buf[0] & GEN_LEN_MASK
                buf = buf[1:]
            elif hdr is Header.EXT_13:
                self._bytes_remaining = ((buf[0] & EXT_13_BYTE0_MASK) << 8) + buf[1]
                buf = buf[2:]
            elif hdr is Header.EXT_16:
                self._bytes_remaining = (buf[1] << 8) + buf[2]
                buf = buf[3:]

        # Append payload to buffer and update remaining / complete
        self._raw_packet.extend(buf)  # type: ignore
        self._bytes_remaining -= len(buf)

        if self._bytes_remaining < 0:
            logger.error("received too much data. parsing is in unknown state")
        elif self._bytes_remaining == 0:
            self._state = GoProResp.State.ACCUMULATED
            self._parse()

    def _parse(self) -> None:
        """Parse the accumulated response (either from a BLE bytestream or an HTTP JSON dict)."""
        # Is this a BLE response?
        if isinstance(self._raw_packet, bytearray):
            buf: bytearray = self._raw_packet  # type: ignore

            # UUID's whose responses contain multiple parameters to parse
            if self.uuid in [UUID.CQ_QUERY_RESP, UUID.CQ_SETTINGS_RESP]:
                identifier: Optional[Union[Type[SettingId], Type[StatusId]]] = None
                if self.uuid is UUID.CQ_SETTINGS_RESP:
                    self._info.append(SettingId(buf[0]))
                    identifier = SettingId
                else:
                    self._info.append(QueryCmdId(buf[0]))
                    if self.id in [
                        QueryCmdId.GET_SETTING_VAL,
                        QueryCmdId.REG_SETTING_VAL_UPDATE,
                        QueryCmdId.SETTING_VAL_PUSH,
                        QueryCmdId.SETTING_CAPABILITY_PUSH,
                        QueryCmdId.GET_CAPABILITIES_VAL,
                        QueryCmdId.REG_CAPABILITIES_UPDATE,
                    ]:
                        identifier = SettingId
                    elif self.id is QueryCmdId.GET_SETTING_NAME:
                        raise NotImplementedError
                    elif self.id in [
                        QueryCmdId.GET_STATUS_VAL,
                        QueryCmdId.REG_STATUS_VAL_UPDATE,
                        QueryCmdId.STATUS_VAL_PUSH,
                    ]:
                        identifier = StatusId

                if identifier is None:
                    raise Exception("Identifier is unexpectedly None")

                # Parse all parameters
                self.status = ErrorCode(buf[1])
                buf = buf[2:]
                while len(buf) != 0:
                    param_id = identifier(buf[0])
                    param_len = buf[1]
                    buf = buf[2:]

                    # Special case where we register for a push notification for something that does not yet
                    # have a value
                    if param_len == 0:
                        self.data[param_id] = []
                        continue

                    param_val = buf[:param_len]
                    buf = buf[param_len:]

                    try:
                        # These can be more than 1 value so use a list
                        if self.cmd in [
                            QueryCmdId.GET_CAPABILITIES_VAL,
                            QueryCmdId.REG_CAPABILITIES_UPDATE,
                            QueryCmdId.SETTING_CAPABILITY_PUSH,
                        ]:
                            # Parse using parser from map and append
                            self.data[param_id].append(GoProResp.parser_map[param_id].parse(param_val))  # type: ignore
                        else:
                            # Parse using parser from map and set
                            self.data[param_id] = GoProResp.parser_map[param_id].parse(param_val)  # type: ignore
                    except KeyError:
                        # We don't have defined params for all ID's yet.
                        logger.warning(f"No parser defined for {param_id}")
                        self.data[param_id] = param_val

            else:  # Other UUID's have responses that can be parsed monolithically
                if self.uuid is UUID.CQ_COMMAND_RESP:
                    self._info.append(CmdId(buf[0]))
                    # If this is a protobuf, first get the action ID (after stripping msb)
                    if self.cmd.value >= 0xF0:  # type: ignore
                        self._info.append(ActionId(buf[1] & 0x7F))
                        buf = buf[1:]
                    buf = buf[1:]

                self.data = GoProResp.parser_map[self.id].parse(buf)  # type: ignore
                self.status = ErrorCode(int(self.data["status"])) if "status" in self.data else self.status  # type: ignore
                scrub(self.data, "status")

        # Is this an HTTP response?
        elif isinstance(self._raw_packet, dict):
            json_data: Dict[str, Any] = self._raw_packet  # type: ignore
            # Is there a parser for this? Most of them do not have one.
            try:
                self.data = GoProResp.parser_map[self.id].parse(json_data)  # type: ignore
            except KeyError:
                self.data = json_data
        else:
            raise Exception(
                f"Unexpected data type ({str(type(self._raw_packet))}) when attempting to parse response"
            )

        # Recursively scrub away parsing artifacts
        scrub(self.data, "_io")
        self._state = GoProResp.State.PARSED



# gopro.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Implements top level interface to GoPro module."""

# TODO make scanning / connect more robust

import re
import csv
import time
import enum
import queue
import asyncio
import logging
import threading
from queue import Queue
from binascii import hexlify
from pathlib import Path
from functools import partial
from typing import Any, Dict, Optional, Pattern, Type, Callable, Generic

import wrapt
import requests

from open_gopro import params
from open_gopro.responses import GoProResp
from open_gopro.constants import CmdId, ErrorCode, UUID, StatusId, QueryCmdId, ProducerType
from open_gopro.ble_commands import BleSettings, BleStatuses, BleCommands, BLECommunicator
from open_gopro.ble_controller import BleakController
from open_gopro.services import AttributeTable, get_gopro_desc
from open_gopro.interfaces import (
    ResponseTimeout,
    ScanFailedToFindDevice,
    ConnectFailed,
    GoProNotInitialized,
    BLEController,
    BleDevice,
    BleClient,
    WifiController,
)
from open_gopro.wifi_commands import WifiCommands, WifiSettings, WifiCommunicator
from open_gopro.wifi_controller import Wireless
from open_gopro.util import SnapshotQueue

logger = logging.getLogger(__name__)

# Send a keep alive minimum every 5 seconds
KEEP_ALIVE_INTERVAL = 60

NUM_THREADS = 3

WRITE_TIMEOUT = 10


class InternalState(enum.IntFlag):
    """State used to manage whether the GoPro instance is ready or not."""

    READY = 0
    ENCODING = 1 << 0
    SYSTEM_BUSY = 1 << 1


@wrapt.decorator
def _ensure_initialized_acquire_ready_semaphore(wrapped, instance, args, kwargs) -> Callable:  # type: ignore
    """If the instance is initialized, acquire the semaphore before doing anything.

    Raises:
        GoProNotInitialized: The function can't be used yet because the GoPro device isn't initialized

    Returns:
        Callable: Function to call after semaphore has been acquired
    """
    if not instance.is_initialized:
        raise GoProNotInitialized
    # Since we're guaranteed to be initialized now, always acquire ready semaphore
    logger.debug(f"{wrapped.__name__} acquiring semaphore")
    with instance.ready:
        logger.debug(f"{wrapped.__name__} has the semaphore")
        ret = wrapped(*args, **kwargs)

    logger.debug(f"{wrapped.__name__} released the semaphore")
    return ret


class GoPro(BLECommunicator, WifiCommunicator, Generic[BleDevice, BleClient]):
    """The top-level BLE and Wifi interface to a GoPro device.

    This will handle for BLE:

    - discovering devices
    - establishing connections
    - discovering GATT characteristics,
    - transferring data

    This will handle for WiFi:

    - finding SSID and password
    - establishing WiFi
    - transferring data

    It will also do some synchronization, etc:

    - ensuring camera is ready / not encoding before transferring data
    - send keep alive periodically

    If no token arg is passed in, the first discovered GoPro device will be connected to.

    It can be used via context manager:

    >>> with GoPro("1234") as gopro:
    >>>     gopro.ble_command.set_shutter(params.Shutter.ON)

    Or without:

    >>> gopro = GoPro("1234")
    >>> gopro.start()
    >>> gopro.ble_command.set_shutter(params.Shutter.ON)

    Args:
        identifier (str, optional): Last 4 of camera name / serial number (i.e. 0456 for GoPro0456). Defaults to None
        ble_adapter (BLEController, optional): Class used to control BLE connection / send data. Defaults to BleakController().
        wifi_adapter (WifiController, optional): Class used to control WiFi connection / send data. Defaults to Wireless().
    """

    base_url = "http://10.5.5.9:8080/"  #: Hard-coded Open GoPro base URL

    def __init__(
        self,
        identifier: Optional[str] = None,
        ble_adapter: Type[BLEController] = BleakController,
        wifi_adapter: Type[WifiController] = Wireless,
    ) -> None:
        # Initialize adapters
        self.ble: BLEController = ble_adapter()
        self.wifi: WifiController = wifi_adapter()

        # Connection information
        self._token: Pattern = re.compile(r"GoPro \d\d\d\d" if identifier is None else f"GoPro {identifier}")
        self.ssid: str
        self.password: str
        self._device: Optional[BleDevice] = None
        self._client: Optional[BleClient] = None
        self.gatt_table: Optional[AttributeTable] = None  #: Attribute storage / access

        # Delegates
        self.ble_command: BleCommands = BleCommands(self)  #: All BLE commands
        self.ble_setting: BleSettings = BleSettings(self)  #: All BLE settings
        self.ble_status: BleStatuses = BleStatuses(self)  #: All BLE statuses
        self.wifi_command: WifiCommands = WifiCommands(self)  #: All Wifi commands
        self.wifi_setting: WifiSettings = WifiSettings(self)  #: All Wifi settings

        # Current accumulating synchronous responses, indexed by UUID. This assumes there can only be one active response per UUID
        self._active_resp: Dict[UUID, GoProResp] = {}
        # Responses that we are waiting for.
        self._sync_resp_wait_q: SnapshotQueue = SnapshotQueue()
        # Synchronous response that has been parsed and are ready for their sender to receive as the response.
        self._sync_resp_ready_q: SnapshotQueue = SnapshotQueue()

        # For outputting asynchronously received information
        self._out_q: "Queue[GoProResp]" = Queue()
        self._listeners: Dict[ProducerType, bool] = {}

        # Set up threads
        self._start_threads_done = 0

        # Thread to run ble controller asyncio loop (to abstract asyncio from client as well as handle async notifications)
        self._module_loop: asyncio.AbstractEventLoop
        self.module_thread = threading.Thread(daemon=True, target=self._run, name="data")
        self.module_thread.start()
        # Wait for this thread to start
        while self._start_threads_done < 1:
            time.sleep(0.1)

        # Set up thread to send keep alive
        self.keep_alive_thread = threading.Thread(
            target=self._periodic_keep_alive, name="keep_alive", daemon=True
        )

        # Set up thread to block until camera is ready to receive commands
        self.ready = threading.BoundedSemaphore(value=1)
        self._state_condition = threading.Condition()
        self._internal_state = InternalState.ENCODING | InternalState.SYSTEM_BUSY
        self.state_thread = threading.Thread(target=self._maintain_state, name="state", daemon=True)

    # TODO if this fails, we will hang and not exit gracefully.
    def __enter__(self) -> "GoPro":
        """Context manager entrance.

        Raises:
            ScanFailedToFindDevice: Couldn't find a relevant device to connect to

        Returns:
            GoPro: initialized GoPro instance
        """
        device = self.scan()
        if device is None:
            raise ScanFailedToFindDevice

        self._device = device
        self.establish_ble(self._device)
        self.discover_chars()
        self.initialize()
        self.establish_wifi()
        return self

    def __exit__(self, *_: Any) -> None:
        self.terminate_wifi()
        self.terminate_ble()

    @property
    def identifier(self) -> Optional[str]:
        """Get a unique identifier for this instance.

        The identifier is the last 4 digits of the camera. That is, the same string that is used to
        scan for the camera for BLE.

        If no token has been provided and a camera is not yet found, this will be None

        Returns:
            Optional[str]: last 4 digits if available, else None
        """
        return None if self._device is None else str(self._device)

    @property
    def is_discovered(self) -> bool:
        """Have we found a relevant GoPro device to connect to?

        Returns:
            bool: True if yes, False if no
        """
        return self._device is not None

    @property
    def is_ble_connected(self) -> bool:
        """Are we connected via BLE to the GoPro device?

        Returns:
            bool: True if yes, False if no
        """
        return self._client is not None

    @property
    def is_wifi_connected(self) -> bool:
        """Are we connected via WiFi to the GoPro device?

        Returns:
            bool: True if yes, False if no
        """
        raise NotImplementedError

    @property
    def is_initialized(self) -> bool:
        """Can we send BLE and Wifi commands right now?

        Returns:
            bool: True if yes, False if no
        """
        return self._start_threads_done >= NUM_THREADS

    def _maintain_state(self) -> None:
        """Thread to keep track of ready / encoding and acquire / release ready semaphore."""
        self.ready.acquire()

        while self.is_ble_connected:
            internal_status_previous = self._internal_state

            with self._state_condition:
                self._state_condition.wait()

                # If we were ready but not now we're not, acquire the semaphore
                if internal_status_previous == 0 and self._internal_state != 0:
                    logger.debug("Control acquiring semaphore")
                    self.ready.acquire()
                    logger.debug("Control has semaphore")
                # If we weren't ready but now we are, release the semaphore
                elif internal_status_previous != 0 and self._internal_state == 0:
                    # If this is the first time, mark that we might now be initialized
                    if not self.is_initialized:
                        self._start_threads_done += 1
                    self.ready.release()
                    logger.debug("Control released semaphore")

        self._start_threads_done -= 1

    def _periodic_keep_alive(self) -> None:
        """Thread to periodically send the keep alive message via BLE."""
        while self.is_ble_connected:
            if not self.is_initialized:
                self._start_threads_done += 1

            try:
                if self.keep_alive():
                    time.sleep(KEEP_ALIVE_INTERVAL)
            except Exception:  # pylint: disable=W0703
                # If the connection disconnects while we were trying to send, there can be any number
                # of exceptions. This is expected and this thread will exit on the next while check.
                pass

        self._start_threads_done -= 1
        logger.debug("periodic keep alive thread exiting...")

    def _run(self) -> None:
        """Thread to keep the event loop running to interface with the BLE adapter."""
        # Create loop for this new thread
        self._module_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._module_loop)

        # Run forever
        self._start_threads_done += 1
        self._module_loop.run_forever()

    def _as_coroutine(self, action: Callable, timeout: float = None) -> Any:
        """Run a function as a coroutine in the module thread.

        This will transfer execution of the given partial to the module thread of this instance

        Args:
            action (Callable): function and parameters to run as corouting
            timeout (float, optional): Time to wait for coroutine to return (in seconds). Defaults to None (wait forever).

        Returns:
            Any: Passes return of coroutine through
        """
        # Allow timeout exception to propagate
        return asyncio.run_coroutine_threadsafe(action(), self._module_loop).result(timeout)

    def start(self) -> None:
        """Convenience API to easily perform all initialization commands if not using the context manager.

        That is, scan and find device, establish connection, discover characteristics, establish wifi, initialize
        """
        self.__enter__()

    def register_listener(self, producer: ProducerType) -> None:
        """Register a producer to store notifications from.

        The notifications can be accessed via the get_update() method.

        Args:
            producer (ProducerType): Producer to listen to.
        """
        self._listeners[producer] = True

    def unregister_listener(self, producer: ProducerType) -> None:
        """Unregister a producer in order to stop listening to its notifications.

        Args:
            producer (ProducerType): Producer to stop listening to.
        """
        del self._listeners[producer]

    def get_update(self, timeout: float = None) -> GoProResp:
        """Get a notification that we received from a registered listener.

        If timeout is None, this will block until a notification is received.

        Args:
            timeout (float, optional): Time to wait for a notification before returning. Defaults to None (wait forever)

        Returns:
            GoProResp: Received notification
        """
        return self._out_q.get(timeout=timeout)

    # ----------------------------------------------- BLE Functionality----------------------------------------

    def scan(self, timeout: float = 5, retries: int = 30) -> Optional[BleDevice]:
        """Scan for devices using the token passed into the instance's initialization.

        Args:
            timeout (float, optional): Time to scan for devices (in seconds). Defaults to 5.
            retries (int, optional): Number of retries before giving up. Defaults to 30.

        Returns:
            Optional[BleDevice]:: device if it was discovered, else None
        """
        for retry in range(1, retries):
            try:
                return self._as_coroutine(partial(self.ble.scan, self._token, timeout))
            except ScanFailedToFindDevice:
                logger.warning(f"Failed to find a device in {timeout} seconds. Retrying #{retry}")
        return None

    def establish_ble(self, device: BleDevice, timeout: float = 15, retries: int = 5) -> None:
        """Connect the instance to a device via BLE.

        Args:
            device (BleDevice): Device to connect to
            timeout (float, optional): Time before considering establishment failed. Defaults to 15 seconds.
            retries (int, optional): How many tries to reconnect after failures. Defaults to 5.

        Raises:
            Any BLE connecting, notification enabling, or pairing errors will be passed through
        """
        for retry in range(1, retries):
            try:
                # Establish connection
                self._client = self._as_coroutine(partial(self.ble.connect, device, timeout=timeout))
                # Attempt to pair
                self._as_coroutine(partial(self.ble.pair, self._client))
                # Enable all GATT notifications
                self._as_coroutine(
                    partial(self.ble.enable_notifications, self._client, self.notification_handler)
                )
                return
            # TODO narrow these down and find out why they're failing, is there a way to handle them?
            except Exception as e:  # pylint: disable=broad-except
                logger.warning(f"{repr(e)}. Retrying #{retry}")
        raise ConnectFailed("BLE", timeout, retries)

    def notification_handler(self, handle: int, data: bytes) -> None:
        """Receive notifications from the BLE controller.

        Args:
            handle (int): Attribute handle that notification was received on.
            data (bytes): Bytestream that was received.
        """
        # Convert handle to UUID
        try:
            uuid = self.gatt_table.handle2uuid(handle)  # type: ignore
        except AttributeError:
            # It's possible a non-Open GoPro service (i.e. battery) sends a notification before we have
            # built the attribute. For now, ignore these.
            logger.warning(f"Unhandled notification from handle {handle}")
            return

        logger.debug(f'Received response on {uuid}: {hexlify(data, ":")}')  # type: ignore

        if uuid not in self._active_resp:
            self._active_resp[uuid] = GoProResp(info=[uuid])

        self._active_resp[uuid]._accumulate(data)

        if self._active_resp[uuid].is_received:
            response = self._active_resp[uuid]

            # Handle internal statuses
            if (
                response.cmd
                in [QueryCmdId.REG_STATUS_VAL_UPDATE, QueryCmdId.GET_STATUS_VAL, QueryCmdId.STATUS_VAL_PUSH]
                and StatusId.ENCODING in response.data
            ):
                with self._state_condition:
                    if response[StatusId.ENCODING] is True:
                        self._internal_state |= InternalState.ENCODING
                    else:
                        self._internal_state &= ~InternalState.ENCODING
                    self._state_condition.notify()
            if (
                response.cmd
                in [QueryCmdId.REG_STATUS_VAL_UPDATE, QueryCmdId.GET_STATUS_VAL, QueryCmdId.STATUS_VAL_PUSH]
                and StatusId.SYSTEM_READY in response.data
            ):
                with self._state_condition:
                    if response[StatusId.SYSTEM_READY] is True:
                        self._internal_state &= ~InternalState.SYSTEM_BUSY
                    else:
                        self._internal_state |= InternalState.SYSTEM_BUSY
                    self._state_condition.notify()

            # Check if this is the awaited synchronous response (id matches). Note! these have to come in order.
            response_claimed = False
            if not self._sync_resp_wait_q.empty():
                queue_snapshot = self._sync_resp_wait_q.snapshot()
                if queue_snapshot[0].id is response.id:
                    # Dequeue it and put this on the ready queue
                    self._sync_resp_wait_q.get_nowait()
                    self._sync_resp_ready_q.put(response)
                    response_claimed = True

            # If this wasn't the awaited synchronous response...
            if not response_claimed:
                logger.info(f"--(ASYNC)--> {response}")
                # See if there are any registered responses that need to be enqueued for client consumption
                for key in response.data.keys():
                    if (response.cmd, key) not in self._listeners:
                        del response.data[key]

                # Enqueue the response if there is anything left
                if len(response.data) > 0:
                    self._out_q.put(response)

            # Clear active response
            del self._active_resp[uuid]

    def terminate_ble(self) -> None:
        """Terminate the BLE connection."""
        logger.info("Terminating the BLE connection")
        self._as_coroutine(partial(self.ble.disconnect, self._client))
        self._client = None

        # try:
        #     self.ready.release()
        # except ValueError:
        #     pass  # If we didn't actually have the semaphore
        # self.keep_alive_thread.join()

        # with self._state_condition:
        #     self._state_condition.notify()
        # self.state_thread.join()

    def discover_chars(self) -> None:
        """Discover all characteristics.

        They will be stored in the instance's gatt_table attribute.
        """
        self.gatt_table = self._as_coroutine(partial(self.ble.discover_chars, self._client))

    def initialize(self) -> None:
        """Register for statuses / information that is necessary for internal functionality."""
        self.state_thread.start()
        self.ble_status.encoding_active.register_value_update()
        self.ble_status.system_ready.register_value_update()
        self.password = self.ble_command.get_wifi_password().flatten
        self.ssid = self.ble_command.get_wifi_ssid().flatten
        self.ble_command.enable_wifi_ap(True)
        self.keep_alive_thread.start()

    def write(self, uuid: UUID, data: bytearray) -> GoProResp:
        """Perform a BLE write and wait for a corresponding notification response.

        There should hopefully not be a scenario where this needs to be called directly as it is generally
        called from the instance's delegates (i.e. self.ble_command, self.ble_setting, self.ble_status)

        Args:
            uuid (UUID): UUID to write to
            data (bytearray): data to write

        Raises:
            Exception: Unexpected functionality occurred

        Returns:
            GoProResp: parsed notification response data
        """
        # Acquire ready semaphore unless this is during initialization or is a Set Shutter Off command
        have_semaphore = False
        if self.is_initialized and not (
            GoProResp.from_write_command(uuid, data).id is CmdId.SET_SHUTTER and data[-1] == 0
        ):
            logger.debug(f"{GoProResp.from_write_command(uuid, data).id} acquiring semaphore")
            self.ready.acquire()
            logger.debug(f"{GoProResp.from_write_command(uuid, data).id} has semaphore")
            have_semaphore = True

        # Store information on the response we are expecting
        self._sync_resp_wait_q.put(GoProResp.from_write_command(uuid, data))
        # Perform write
        self._as_coroutine(partial(self.ble.write, self._client, uuid.value, data))
        # Wait to be notified that response was received
        try:
            response = self._sync_resp_ready_q.get(timeout=WRITE_TIMEOUT)
        except queue.Empty as e:
            logger.error(f"Response timeout of {WRITE_TIMEOUT} seconds!")
            raise ResponseTimeout(WRITE_TIMEOUT) from e

        # Check status
        try:
            if response.status is not ErrorCode.SUCCESS:
                logger.warning(f"Received non-success status: {response.status}")
        except AttributeError:
            logger.error("Not able to parse status from response")

        # If this was set shutter on, we need to wait to be notified that encoding has started
        if response.cmd is CmdId.SET_SHUTTER and data[-1] == 1:
            while not self._internal_state & InternalState.ENCODING:
                # We don't want to use the application's loop, can't use any of our loops due to potential deadlock,
                # and don't want to spawn a new thread for this. So just poll ¯\_(ツ)_/¯
                # A read to an int is atomic anyway.
                time.sleep(0.1)

        # Release the semaphore if we acquired it
        if have_semaphore:
            self.ready.release()
            logger.debug(f"{GoProResp.from_write_command(uuid, data).id} released the semaphore")

        if response is None:
            raise Exception("The synchronous response is unexpectedly None")

        return response

    def read(self, uuid: UUID) -> GoProResp:
        """Read a characteristic's data by UUID.

        There should hopefully not be a scenario where this needs to be called directly as it is generally
        called from the instance's delegates (i.e. self.command, self.setting, self.ble_status)

        Args:
            uuid (UUID): characteristic data to read

        Returns:
            bytearray: read data
        """
        have_semaphore = False
        if self.is_initialized:
            logger.debug(f"{uuid} acquiring semaphore")
            self.ready.acquire()
            logger.debug(f"{uuid} has the semaphore")
            have_semaphore = True

        received_data = self._as_coroutine(partial(self.ble.read, self._client, uuid.value))

        if have_semaphore:
            self.ready.release()
            logger.debug(f"{uuid} released the semaphore")

        return GoProResp.from_read_response(uuid, received_data)

    def services_as_csv(self, file: Path = Path("gopro_services.csv")) -> None:
        """Dump discovered services to a csv file.

        If the services haven't been discovered, they will be now.

        Args:
            file (Path, optional): File to write to. Defaults to Path("gopro_services.csv").
        """
        # If we don't yet have the characteristic information, perform a service discovery
        if self.gatt_table is None:
            self.discover_chars()
        with open(file, mode="w") as f:
            logger.debug(f"Dumping discovered BLE characteristics to {file}")
            w = csv.writer(f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL)
            w.writerow(["handle", "description", "UUID", "properties", "value"])
            # For each service in table
            for s in self.gatt_table.services.values():  # type: ignore
                desc = get_gopro_desc(s.uuid.value) if s.name.lower() == "unknown" else s.name
                w.writerow(["SERVICE", desc, s.uuid.value, "SERVICE", "SERVICE"])
                # For each characteristic in service
                for c in s.chars.values():
                    desc = get_gopro_desc(c.uuid.value) if c.name.lower() == "unknown" else c.name
                    w.writerow([c.handle, desc, c.uuid.value, " ".join(c.props), c.value])
                    # For each descriptor in characteristic
                    for d in c.descriptors:
                        w.writerow([d.handle, "DESCRIPTOR", "", "", d.value])

    # ---------------------------------WiFi-------------------------------------------------------------------

    def establish_wifi(self, timeout: float = 15, retries: int = 5) -> None:
        """Connect to a GoPro device via WiFi.

        Args:
            timeout (float, optional): Time before considering establishment failed. Defaults to 15 seconds.
            retries (int, optional): How many tries to reconnect after failures. Defaults to 5.

        Raises:
            Exception: Wifi failed to connect.
        """
        logger.info("Establishing WiFi connection...")
        for retry in range(1, retries):
            if self.wifi.connect(self.ssid, self.password, timeout=timeout):
                return
            logger.warning(f"WiFi connection timed out. Retrying #{retry}")
        raise ConnectFailed("WiFi", timeout, retries)

    def terminate_wifi(self) -> None:
        """Terminate the WiFi connection."""
        logger.debug("Terminating the WiFi connection...")
        self.wifi.disconnect()

    @_ensure_initialized_acquire_ready_semaphore
    def get(self, url: str) -> GoProResp:
        """Send an HTTP GET request to an Open GoPro endpoint.

        There should hopefully not be a scenario where this needs to be called directly as it is generally
        called from the instance's delegates (i.e. self.wifi_command and self.wifi_status)

        Args:
            url (str): endpoint URL

        Raises:
            requests.models.HTTPError if there is an HTTP error

        Returns:
            GoProResp: response
        """
        url = GoPro.base_url + url
        logger.debug(f"Sending:  {url}")

        # TODO This is a terrible temporary hack. I just don't know why these are breaking.
        time.sleep(2)

        request = requests.get(url)
        request.raise_for_status()
        response = GoProResp.from_http_response(request)
        return response

    @_ensure_initialized_acquire_ready_semaphore
    def stream_to_file(self, url: str, file: Path) -> None:
        """Send an HTTP GET request to an Open GoPro endpoint to download a binary file.

        There should hopefully not be a scenario where this needs to be called directly as it is generally
        called from the instance's delegates (i.e. self.wifi_command and self.wifi_status)

        Args:
            url (str): endpoint URL
            file (Path): location where file should be downloaded to
        """
        url = GoPro.base_url + url
        logger.debug(f"Sending: {url}")
        with requests.get(url, stream=True) as request:
            request.raise_for_status()
            with open(file, "wb") as f:
                logger.debug(f"receiving stream to {file}...")
                for chunk in request.iter_content(chunk_size=8192):
                    f.write(chunk)

    # -----------------------------------Abstracted Data Functionality----------------------------------------

    def keep_alive(self) -> bool:
        """Send a heartbeat to prevent the BLE connection from dropping.

        This is sent automatically by the GoPro instance.

        Returns:
            bool: True if it succeeded,. False otherwise
        """
        return self.ble_setting.led.set(params.LED.BLE_KEEP_ALIVE).is_ok



# ble_controller.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Manage a Bluetooth connection using bleak."""

import asyncio
import logging
from typing import Pattern, Dict, Any

from bleak import BleakScanner, BleakClient
from bleak.backends.device import BLEDevice as BleakDevice

from open_gopro.constants import UUID
from open_gopro.util import Singleton
from open_gopro.services import Service, Characteristic, Descriptor, AttributeTable
from open_gopro.interfaces import ScanFailedToFindDevice, BLEController, NotiHandlerType

logger = logging.getLogger(__name__)


class Client:
    """Information per client to be stored in class database."""

    def __init__(self) -> None:
        self.disconnect_event = asyncio.Event()

    def disconnect_handler(self, client: BleakClient) -> None:
        """Handle the disconnect callback from bleak.

        Args:
            client (BleakClient): connected client to disconnect from
        """
        logger.debug(f"On device {client}, Disconnected callback called!")
        self.disconnect_event.set()


class BleakController(BLEController, Singleton):
    """Wrapper around bleak to manage a Bluetooth connection.

    Note, this is a singleton.
    """

    clients: Dict[BleakClient, Client] = {}

    async def read(self, client: BleakClient, uuid: str) -> bytearray:
        """AsyncIO method to read data from a UUID.

        Args:
            client (BleakClient): client to read from
            uuid (str): uuid to read

        Returns:
            bytearray: read data
        """
        logger.debug(f"Reading from {uuid}")
        response = await client.read_gatt_char(uuid)
        logger.debug(f'Received response on {uuid}: {response.hex( ":")}')

        return response

    async def write(self, client: BleakClient, uuid: str, value: bytearray) -> None:
        """AsyncIO method to write data to a UUID.

        Perform a write, then wait for event to be notified (from a notification handler)

        Args:
            client (BleakClient): Device to write to
            uuid (str): characteristic UUID to write to
            value (bytearray): data to write
            event (asyncio.Event): event to synchronize response
        """
        logger.debug(f"Writing to {uuid}: {value.hex(':')}")

        await client.write_gatt_char(uuid, value)

    async def scan(self, token: Pattern, timeout: int = 5) -> BleakDevice:
        """AsyncIO method to scan for a regex.

        Args:
            token (Pattern): Regex to look for when scanning.
            timeout (int, optional): Time to scan. Defaults to 5.

        Raises:
            ScanFailedToFindDevice: Did not find any of the token when scanning.

        Returns:
            BleakDevice: The first matched device that was discovered
        """
        name = token.pattern.strip(r"\d\d\d\d")
        logger.info(f"Scanning for {name} bluetooth devices...")

        devices: Dict[str, BleakDevice] = {}

        def _scan_callback(device: BleakDevice, _: Any) -> None:
            """Bleak optional scan callback to receive every scan result.

            We need this because the GoPro will sometimes only show as a
            nonconnectable scan response from bleak.

            Also, if the device shows as both the above as well as a connectable advertisement,
            the latter must be used in order to avoid an exception from CoreBluetooth on Mac.

            Args:
                device (BleakDevice): discovered device
                _ : advertisement that we're ignoring
            """
            # Add to the dict if not unknown
            if device.name != "Unknown" and device.name is not None:
                devices[device.name] = device

        # Now get list of connectable advertisements
        for device in await BleakScanner.discover(timeout=timeout, detection_callback=_scan_callback):
            if device.name != "Unknown" and device.name is not None:
                devices[device.name] = device

        for d in devices:
            logger.info(f"\tDiscovered: {d}")

        # Now look for our matching device(s)
        matched_devices = [device for name, device in devices.items() if token.match(name)]
        if len(matched_devices) == 0:
            raise ScanFailedToFindDevice
        logger.info(f"Found {len(matched_devices)} matching devices.")

        # If there's more than 1, the first one gets lucky.
        return matched_devices[0]

    async def connect(self, device: BleakDevice, timeout: int = 15) -> BleakClient:
        """AsyncIO method to connect to a device.

        Args:
            device (BleakDevice): Device to connect to

        Raises:
            Exception: Connection failed to establish

        Returns:
            BleakClient: Connected device
        """
        c = Client()
        bleak_client = BleakClient(device, disconnected_callback=c.disconnect_handler)
        BleakController.clients[bleak_client] = c

        logger.info(f"Establishing BLE connection to {device}...")
        await bleak_client.connect(timeout=timeout)

        return bleak_client

    async def pair(self, client: BleakClient) -> None:
        """AsyncIO method to pair to a device after connection.

        This is required for Windows and not allowed on Mac...

        Args:
            client (BleakClient): Device to pair to
        """
        logger.debug("Attempting to pair...")
        try:
            await client.pair()
        except NotImplementedError:
            # This is expected on Mac
            pass
        logger.debug("Pairing complete!")
        logger.info("BLE Connected!")

    async def enable_notifications(self, client: BleakClient, handler: NotiHandlerType) -> None:
        """AsyncIO method to enable all notifications.

        Search through all characteristics and enable any that have notification property.

        Args:
            client (BleakClient): Device to enable notifications for
            handler (Callable): Notification callback handler
        """
        logger.info("Enabling notifications...")
        for service in client.services:
            for char in service.characteristics:
                if "notify" in char.properties:
                    logger.debug(f"Enabling notification on char {char.uuid}")
                    await client.start_notify(char, handler)
        logger.info("Done enabling notifications")

    async def discover_chars(self, client: BleakClient) -> AttributeTable:
        """AsyncIO method to discover all characteristics.

        Args:
            client (BleakClient): Device to perform discovery on

        Returns:
            AttributeTable: Dict of services by UUID
        """
        logger.info("Discovering characteristics...")
        services: Dict[UUID, Service] = {}

        for service in client.services:
            logger.debug("[Service] {0}: {1}".format(service.uuid, service.description))
            try:
                # Create new service
                services[UUID(service.uuid)] = Service(UUID(service.uuid), service.description)
            except ValueError:
                logger.error(f"{service.uuid} is not a known service")
                continue

            # Loop over all chars in service
            chars: Dict[UUID, Characteristic] = {}
            for char in service.characteristics:
                # Read if applicable
                value = bytes(await client.read_gatt_char(char.uuid)) if "read" in char.properties else b""

                # Create and log characteristic
                c = Characteristic(char.handle, UUID(char.uuid), char.properties, char.description, value)

                # Get any descriptors if they exist
                descriptors = []
                for descriptor in char.descriptors:
                    value = await client.read_gatt_descriptor(descriptor.handle)
                    descriptors.append(Descriptor(descriptor.handle, value))

                # Update char and add to dict
                c.descriptors = descriptors
                # Add characteristic to char dic
                chars[c.uuid] = c

            # Add char dict to service
            services[UUID(service.uuid)].chars = chars

        logger.info("Done discovering characteristics!")
        return AttributeTable(services)

    async def disconnect(self, client: BleakClient) -> bool:
        """AsyncIO method to terminate a BLE connection.

        Args:
            client (BleakClient): client to disconnect from

        Returns:
            bool: True if disconnect was successful, False otherwise
        """
        if client.is_connected:
            logger.info("Disconnecting...")
            await client.disconnect()
            await BleakController.clients[client].disconnect_event.wait()
        logger.info("Device disconnected!")

        return True



# __init__.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""All GoPro parameters (both defined constants and generated from protobufs)."""

from open_gopro.proto.request_get_preset_status_pb import EnumRegisterPresetStatus
from open_gopro.params.params import *

__all__ = ["EnumRegisterPresetStatus"]



# __init__.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""All GoPro exports that the the user will want should be exported here."""

from open_gopro import params, constants
from open_gopro.gopro import GoPro
from open_gopro.constants import UUID, SettingId, StatusId, CmdId, QueryCmdId
from open_gopro.responses import GoProResp

__all__ = [
    "params",
    "GoPro",
    "UUID",
    "SettingId",
    "StatusId",
    "CmdId",
    "GoProResp",
    "constants",
    "QueryCmdId",
]



# turbo_transfer_pb.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: turbo_transfer.proto
# plugin: python-betterproto
from dataclasses import dataclass

import betterproto


@dataclass
class RequestSetTurboActive(betterproto.Message):
    active: bool = betterproto.bool_field(1)
